import atexit
import os

import pytest
from PIL import Image, UnidentifiedImageError

from term_image.exceptions import URLNotFoundError
from term_image.image import BaseImage, BlockImage, ImageSource, Size, from_url

python_image = "tests/images/python.png"
python_url = (
    "https://raw.githubusercontent.com/AnonymouX47/term-image/main/tests/"
    "images/python.png"
)
python_img = Image.open(python_image)


@atexit.register
def close_imgs():
    python_img.close()


def test_from_url():
    with pytest.raises(TypeError, match=r"'url'"):
        BlockImage.from_url(python_img)
    with pytest.raises(ValueError, match="Invalid URL"):
        BlockImage.from_url(python_image)
    with pytest.raises(URLNotFoundError):
        BlockImage.from_url(python_url + "e")
    with pytest.raises(UnidentifiedImageError):
        BlockImage.from_url(
            "https://raw.githubusercontent.com/AnonymouX47/term-image/main/LICENSE"
        )

    image = BlockImage.from_url(python_url)
    assert isinstance(image, BlockImage)
    assert image._url == python_url
    assert os.path.exists(image._source)
    assert image._source_type is ImageSource.URL

    # Ensure size arguments get through
    with pytest.raises(TypeError, match="'width' and 'height'"):
        BlockImage.from_url(python_url, width=1, height=Size.FIT)


def test_source():
    image = BlockImage.from_url(python_url)
    assert image.source == image._url == python_url
    assert image.source_type is ImageSource.URL


def test_close():
    image = BlockImage.from_url(python_url)
    source = image._source

    image.close()
    assert not os.path.exists(source)
    with pytest.raises(AttributeError):
        image._url


class TestFactoryFunction:
    def test_from_url(self):
        with pytest.raises(TypeError, match=r"'url'"):
            from_url(python_img)

        # Ensure size arguments get through
        with pytest.raises(TypeError, match="'width' and 'height'"):
            from_url(python_url, width=1, height=Size.FIT)

        assert isinstance(from_url(python_url), BaseImage)



import os
from contextlib import contextmanager

import term_image


def get_cell_size():
    return cell_size


def set_cell_size(size):
    global cell_size

    cell_size = size
    term_image.AutoCellRatio.is_supported = None


@contextmanager
def reset_cell_size_ratio():
    cell_ratio = term_image._cell_ratio
    cell_size = get_cell_size()
    try:
        yield
    finally:
        term_image._cell_ratio = cell_ratio
        set_cell_size(cell_size)


def get_terminal_size():
    return os.terminal_size((80, 30))


def get_terminal_name_version():
    return terminal_name_version


def set_terminal_name_version(name: str, version: str = ""):
    global terminal_name_version

    terminal_name_version = (name, version)


def get_fg_bg_colors(*, hex=False):
    return (
        tuple(rgb and "#" + "".join(f"{x:02x}" for x in rgb) for rgb in fg_bg)
        if hex
        else fg_bg
    )


def set_fg_bg_colors(fg=None, bg=None):
    global fg_bg
    fg_bg = (fg, bg)


def _is_on_kitty():
    return is_on_kitty


def toggle_is_on_kitty():
    global is_on_kitty
    is_on_kitty = not is_on_kitty


term_image.utils.get_terminal_size = get_terminal_size

terminal_name_version = ("", "")
term_image.utils.get_terminal_name_version = get_terminal_name_version

cell_size = None
term_image.get_cell_size = get_cell_size
term_image.utils.get_cell_size = get_cell_size

fg_bg = [(0, 0, 0), (0, 0, 0)]
term_image.utils.get_fg_bg_colors = get_fg_bg_colors

import term_image.image  # noqa: E402

term_image.image.GraphicsImage._supported = True

is_on_kitty = False
term_image.image.TextImage._is_on_kitty = staticmethod(_is_on_kitty)



import gc
import io
import sys
from contextlib import contextmanager

import pytest
import urwid
from PIL import Image

from term_image import ctlseqs
from term_image.exceptions import RenderError, UrwidImageError
from term_image.image import (
    BlockImage,
    GraphicsImage,
    ITerm2Image,
    KittyImage,
    Size,
    TextImage,
)
from term_image.image.common import _ALPHA_THRESHOLD
from term_image.widget import UrwidImage, UrwidImageCanvas, UrwidImageScreen

from .. import get_terminal_name_version, set_terminal_name_version

_size = (30, 15)

python_file = "tests/images/python.png"
python_img = Image.open(python_file)
python_image = BlockImage(python_img)
kitty_image = KittyImage(python_img)
trans = BlockImage.from_file("tests/images/trans.png")


class TestWidget:
    def test_args(self):
        for value in (python_file, python_img):
            with pytest.raises(TypeError, match="'image'"):
                UrwidImage(value)

        for value in (None, 2.0, 2):
            with pytest.raises(TypeError, match="'format_spec'"):
                UrwidImage(python_image, value)
        for string in (".", "+"):
            with pytest.raises(ValueError, match="format specifier"):
                UrwidImage(python_image, string)

        for value in (None, 2.0, "2"):
            with pytest.raises(TypeError, match="'upscale'"):
                UrwidImage(python_image, upscale=value)

    def test_attributes(self):
        image_w = UrwidImage(python_image)
        assert all(name.startswith("_ti_") for name in vars(image_w))

    def test_image(self):
        image_w = UrwidImage(python_image)
        assert image_w.image is python_image

    def test_error_placeholder(self):
        image = BlockImage(Image.new("La", (1, 1)))
        image_w = UrwidImage(image)
        placeholder = urwid.SolidFill("?")

        with pytest.raises(RenderError, match="convert"):
            image_w.render(_size)

        image_w.set_error_placeholder(placeholder)
        canv = image_w.render(_size)
        assert isinstance(canv, urwid.CompositeCanvas)
        assert len(canv.shards) == 1

        solid_canv = canv.shards[0][1][0][-1]
        assert isinstance(solid_canv, urwid.SolidCanvas)
        assert solid_canv.widget_info[0] is placeholder
        assert solid_canv.size == _size


class TestRows:
    def test_upscale_false(self):
        image_w = UrwidImage(python_image)
        ori_size = python_image._valid_size(Size.ORIGINAL)
        fit_size = python_image._valid_size(ori_size[0] - 5)

        assert ori_size[1] != fit_size[1]
        assert image_w.rows((ori_size[0] - 5,)) == fit_size[1]
        assert image_w.rows((ori_size[0] + 5,)) == ori_size[1]

    def test_upscale_true(self):
        image_w = UrwidImage(python_image, upscale=True)
        ori_size = python_image._valid_size(Size.ORIGINAL)

        assert (
            image_w.rows((ori_size[0] - 5,))
            == image_w._ti_image._valid_size(ori_size[0] - 5)[1]
            != ori_size[1]
        )
        assert (
            image_w.rows((ori_size[0] + 5,))
            == image_w._ti_image._valid_size(ori_size[0] + 5)[1]
            != ori_size[1]
        )


class TestRender:
    class TestBox:
        def _test_output(self, canv, size, image_size):
            trans._size = image_size
            render = trans._format_render(
                trans._renderer(
                    trans._render_image, _ALPHA_THRESHOLD, split_cells=True
                ),
                None,
                size[0],
                None,
                size[1],
            ).splitlines()
            lines = canv._ti_lines

            assert canv.size == size
            assert len(lines) == size[1] == len(render)
            for line, render_line in zip(lines, render):
                assert isinstance(line, bytes)
                assert line.decode() == render_line + "\0\0"

        def test_smaller_than_original_size(self):
            not_upscaled = UrwidImage(trans)
            upscaled = UrwidImage(trans, upscale=True)
            ori_size = trans._valid_size(Size.ORIGINAL)
            size = tuple((x - 2 for x in ori_size))
            upscaled_canv = upscaled.render(size)
            not_upscaled_canv = not_upscaled.render(size)
            image_size = trans._valid_size(Size.FIT, frame_size=size)

            self._test_output(upscaled_canv, size, image_size)
            self._test_output(not_upscaled_canv, size, image_size)
            assert upscaled_canv._ti_lines == not_upscaled_canv._ti_lines

        def test_upscale_false(self):
            image_w = UrwidImage(trans)
            ori_size = trans._valid_size(Size.ORIGINAL)
            size = tuple((x + 2 for x in ori_size))
            canv = image_w.render(size)
            lines = canv._ti_lines

            self._test_output(canv, size, ori_size)

            # One padding line at both the top and bottom
            assert lines[0].rstrip(b"\0\0").isspace()
            assert not lines[1].rstrip(b"\0\0").isspace()
            assert not lines[-2].rstrip(b"\0\0").isspace()
            assert lines[-1].rstrip(b"\0\0").isspace()

            # One padding column on both ends of each line
            for line in lines:
                line = line.decode().rstrip("\0\0")
                if not line.isspace():
                    left, _, right = line.split(ctlseqs.SGR_NORMAL)
                    assert left == " "
                    assert right == " "

        def test_upscale_true(self):
            image_w = UrwidImage(trans, upscale=True)
            ori_size = trans._valid_size(Size.ORIGINAL)
            size = trans._valid_size(ori_size[0] + 2)
            canv = image_w.render(size)
            lines = canv._ti_lines
            image_size = trans._valid_size(Size.FIT, frame_size=size)

            # The image should have no padding; *size* is proportional to image size
            # and sizing is FIT.

            self._test_output(canv, size, image_size)
            assert not lines[0].rstrip(b"\0\0").isspace()
            assert not lines[-1].rstrip(b"\0\0").isspace()
            for line in lines:
                line = line.decode().rstrip("\0\0")
                if not line.isspace():
                    left, _, right = line.split(ctlseqs.SGR_NORMAL)
                    assert left == ""
                    assert right == ""

    class TestFlow:
        def _test_output(self, canv, size, image_size, upscaled):
            trans._size = image_size
            render = trans._format_render(
                trans._renderer(
                    trans._render_image, _ALPHA_THRESHOLD, split_cells=True
                ),
                None,
                size[0],
                None,
                1,
            ).splitlines()
            lines = canv._ti_lines

            assert canv.size == (size[0], image_size[1])
            assert len(lines) == image_size[1] == len(render)
            if upscaled:
                assert size[0] == image_size[0]
            for line, render_line in zip(lines, render):
                assert isinstance(line, bytes)
                assert line.decode() == render_line + "\0\0"

        def test_smaller_than_original_size(self):
            not_upscaled = UrwidImage(trans)
            upscaled = UrwidImage(trans, upscale=True)
            ori_size = trans._valid_size(Size.ORIGINAL)
            size = (ori_size[0] - 2,)
            upscaled_canv = upscaled.render(size)
            not_upscaled_canv = not_upscaled.render(size)
            image_size = trans._valid_size(size[0])

            self._test_output(upscaled_canv, size, image_size, True)
            self._test_output(not_upscaled_canv, size, image_size, False)
            assert upscaled_canv._ti_lines == not_upscaled_canv._ti_lines

        def test_upscale_false(self):
            image_w = UrwidImage(trans)
            ori_size = trans._valid_size(Size.ORIGINAL)
            size = (ori_size[0] + 2,)
            canv = image_w.render(size)
            lines = canv._ti_lines

            self._test_output(canv, size, ori_size, False)

            # No vertical padding
            assert not lines[1].rstrip(b"\0\0").isspace()
            assert not lines[-2].rstrip(b"\0\0").isspace()

            # One padding column on both ends of each line
            for line in lines:
                line = line.decode().rstrip("\0\0")
                if not line.isspace():
                    left, _, right = line.split(ctlseqs.SGR_NORMAL)
                    assert left == " "
                    assert right == " "

        def test_upscale_true(self):
            image_w = UrwidImage(trans, upscale=True)
            ori_size = trans._valid_size(Size.ORIGINAL)
            size = (ori_size[0] + 2,)
            canv = image_w.render(size)
            lines = canv._ti_lines

            # The image should have no padding

            self._test_output(canv, size, trans._valid_size(size[0]), True)
            assert not lines[0].rstrip(b"\0\0").isspace()
            assert not lines[-1].rstrip(b"\0\0").isspace()
            for line in lines:
                line = line.decode().rstrip("\0\0")
                if not line.isspace():
                    left, _, right = line.split(ctlseqs.SGR_NORMAL)
                    assert left == ""
                    assert right == ""

    def test_fixed(self):
        image_w = UrwidImage(trans)
        with pytest.raises(UrwidImageError, match="fixed"):
            image_w.render(())


def test_ignore_padding():
    canvases_lines = [
        UrwidImage(python_image, fmt).render(_size)._ti_lines
        for fmt in ("", "200.200", "100.50")
    ]
    python_image.set_size(Size.AUTO, frame_size=_size)
    render = (
        python_image._format_render(
            python_image._renderer(
                python_image._render_image, _ALPHA_THRESHOLD, split_cells=True
            ),
            None,
            _size[0],
            None,
            _size[1],
        )
        .encode()
        .splitlines()
    )

    for render_line, line1, line2, line3 in zip(render, *canvases_lines):
        assert render_line + b"\0\0" == line1 == line2 == line3


# There's no need to test formatting separately since it uses the functionality of the
# underlying image


class TestKittyWidget:
    def test_disguise(self):
        image_w = UrwidImage(kitty_image)
        canv = image_w.render(_size)
        UrwidImageCanvas._ti_disguise_state = 0
        for disguise_state in (0, 1, 2, 0, 1, 2, 0):
            content = canv.text
            for line in content:
                line = line.decode()
                assert line.endswith("\0\0" + "\b " * disguise_state)
            image_w._ti_change_disguise()

    def test_disguise_combined(self):
        image_w = UrwidImage(kitty_image)
        canv = image_w.render(_size)
        UrwidImageCanvas._ti_disguise_state = 2
        try:
            for disguise_states in ((2, 0), (0, 1), (1, 2), (2, 0)):
                content = canv.text
                for line in content:
                    line = line.decode()
                    assert line.endswith("\0\0" + "\b " * sum(disguise_states))
                image_w._ti_change_disguise()
                UrwidImageCanvas._ti_change_disguise()
        finally:
            UrwidImageCanvas._ti_disguise_state = 0

    class TestZIndex:
        def test_alloc(self):
            gc.collect()
            UrwidImage._ti_next_z_index = 1
            UrwidImage._ti_free_z_indexes.clear()
            images = []

            for index in (1, -1, 2, -2, 3, -3, 4, -4):
                images.append(UrwidImage(kitty_image))
                assert images[-1]._ti_z_index == index
            assert UrwidImage._ti_next_z_index == 5

        def test_free(self):
            gc.collect()
            UrwidImage._ti_next_z_index = 1
            UrwidImage._ti_free_z_indexes.clear()

            [UrwidImage(kitty_image) for _ in range(10)]
            gc.collect()

            assert len(UrwidImage._ti_free_z_indexes) == 10
            assert UrwidImage._ti_next_z_index == 6

        def test_use_freed(self):
            gc.collect()
            UrwidImage._ti_next_z_index = 1
            UrwidImage._ti_free_z_indexes.clear()

            images = [UrwidImage(kitty_image) for _ in range(100)]
            z_indexes = {image_w._ti_z_index for image_w in images}
            next_z_index = UrwidImage._ti_next_z_index
            del images
            gc.collect()
            images = [UrwidImage(kitty_image) for _ in range(100)]

            assert {image_w._ti_z_index for image_w in images} == z_indexes
            assert len(UrwidImage._ti_free_z_indexes) == 0
            assert next_z_index == UrwidImage._ti_next_z_index

        def test_limit(self):
            gc.collect()
            UrwidImage._ti_next_z_index = 2**31 - 2
            UrwidImage._ti_free_z_indexes.clear()
            images = []

            for index in (2**31 - 2, -(2**31 - 2), 2**31 - 1, -(2**31 - 1)):
                images.append(UrwidImage(kitty_image))
                assert images[-1]._ti_z_index == index

            with pytest.raises(UrwidImageError, match="many .* kitty"):
                UrwidImage(kitty_image)
            assert UrwidImage._ti_next_z_index == 2**31

            # Does not recover until after a widget is freed
            with pytest.raises(UrwidImageError, match="many .* kitty"):
                UrwidImage(kitty_image)
            assert UrwidImage._ti_next_z_index == 2**31

            del images[0]
            gc.collect()

            # Recovers after a widget is freed
            images.append(UrwidImage(kitty_image))
            assert images[-1]._ti_z_index == 2**31 - 2
            assert UrwidImage._ti_next_z_index == 2**31

            # Fails again after freed index is used
            with pytest.raises(UrwidImageError, match="many .* kitty"):
                UrwidImage(kitty_image)
            assert UrwidImage._ti_next_z_index == 2**31


class TestCanvas:
    def test_cols(self):
        image_w = UrwidImage(python_image, upscale=True)
        canv = image_w.render(_size)
        assert canv.cols() == _size[0]

    def test_rows(self):
        image_w = UrwidImage(python_image, upscale=True)
        canv = image_w.render(_size)
        assert canv.rows() == _size[1]

    def test_content(self):
        image_w = UrwidImage(python_image, upscale=True)
        canv = image_w.render(_size)
        content = canv.text

        assert len(content) == len(canv._ti_lines)
        for content_line, line in zip(content, canv._ti_lines):
            assert content_line == line.replace(b"\0", b"") + b"\0\0"

    def test_disguise_supported(self):
        _terminal_name_version = get_terminal_name_version()
        try:
            set_terminal_name_version("konsole")
            for ImageClass in (KittyImage, ITerm2Image):
                image_w = UrwidImage(ImageClass(python_img))
                canv = image_w.render(_size)

                UrwidImageCanvas._ti_disguise_state = 0
                for disguise_state in (0, 1, 2, 0, 1, 2, 0):
                    content = canv.text
                    for line in content:
                        line = line.decode()
                        assert line.endswith("\0\0" + "\b " * disguise_state)
                    UrwidImageCanvas._ti_change_disguise()
        finally:
            UrwidImageCanvas._ti_disguise_state = 0
            set_terminal_name_version(*_terminal_name_version)

    def test_disguise_not_supported(self):
        _terminal_name_version = get_terminal_name_version()
        try:
            set_terminal_name_version("wezterm")
            for ImageClass in (BlockImage, ITerm2Image):
                image_w = UrwidImage(ImageClass(python_img))
                canv = image_w.render(_size)

                UrwidImageCanvas._ti_disguise_state = 0
                for disguise_state in (0, 1, 2, 0, 1, 2, 0):
                    content = canv.text
                    for line in content:
                        line = line.decode()
                        assert line.endswith("\0\0")
                    UrwidImageCanvas._ti_change_disguise()
        finally:
            UrwidImageCanvas._ti_disguise_state = 0
            set_terminal_name_version(*_terminal_name_version)


def get_trim_render_canv(ImageClass, image_size, h_align, v_align):
    image = ImageClass.from_file("tests/images/trans.png")
    image._size = image_size
    style_args = {"split_cells": True} if issubclass(ImageClass, TextImage) else {}
    render = image._renderer(image._render_image, _ALPHA_THRESHOLD, **style_args)
    image_w = UrwidImage(image, f"{h_align}.{v_align}")
    canv = UrwidImageCanvas(
        image._format_render(render, h_align, _size[0], v_align, _size[1]),
        _size,
        image_size,
    )
    canv._widget_info = (image_w, _size, False)

    pad = _size[0] - image_size[0]
    if h_align == "<":
        pad_left = 0
        pad_right = pad
    elif h_align == ">":
        pad_left = pad
        pad_right = 0
    else:
        pad_left = pad // 2
        pad_right = pad - pad_left

    if issubclass(ImageClass, TextImage):
        pad = _size[1] - image_size[1]
        if v_align == "^":
            pad_top = 0
            pad_bottom = pad
        elif v_align == "_":
            pad_top = pad
            pad_bottom = 0
        else:
            pad_top = pad // 2
            pad_bottom = pad - pad_top

        padding_line = [[b" "] * _size[0]]
        left_padding = [b" "] * pad_left
        right_padding = [b" "] * pad_right

        render = (
            padding_line * pad_top
            + [
                left_padding + line.split(b"\0") + right_padding
                for line in render.encode().split(b"\n")
            ]
            + padding_line * pad_bottom
        )

        return render, canv, pad_top, pad_bottom, pad_left, pad_right
    else:
        render = (
            image._format_render(render, h_align, _size[0], v_align, _size[1])
            .encode()
            .split(b"\n")
        )

        return render, canv


def content_to_text(content_iter):
    return [b"".join(text for *_, text in line) for line in content_iter]


def get_all_trim_args():
    args = (
        (trim_left, trim_top, cols, rows)
        for trim_top in range(0, _size[1], 2)
        for rows in range(_size[1] - trim_top, 0, -2)
        for trim_left in range(0, _size[0], 2)
        for cols in range(_size[0] - trim_left, 0, -2)
        if cols != _size[0]
    )

    return args


def get_hori_trim_args():
    args = (
        (trim_left, cols)
        for trim_left in range(0, _size[0], 2)
        for cols in range(_size[0] - trim_left, 0, -2)
    )
    next(args)  # skip whole canvas

    return args


class TestCanvasTrim:
    class TestCalcTrim:
        # Layout of a formatted render:
        #
        #      A   B      C   D
        #    ->|   |      |   |<-
        #
        # '->' - side1 trim direction
        # A-B  - side1 padding
        # B-C  - image
        # C-D  - side2 padding
        # '<-' - side2 trim direction

        # A=B=0, C=10, D=20
        def test_side1_aligned(self):
            for args, result in (
                # side1 trim at A/B
                ((20, 10, 0, 0, 0, 10), (0, 0, 0, 10)),  # side2 trim at D
                ((20, 10, 0, 0, 2, 10), (0, 0, 0, 8)),  # side2 trim between C and D
                ((20, 10, 0, 0, 10, 10), (0, 0, 0, 0)),  # side2 trim at C
                ((20, 10, 0, 0, 12, 10), (0, 0, 2, 0)),  # side2 trim between B and C
                # side1 trim between B and C
                ((20, 10, 2, 0, 0, 10), (0, 2, 0, 10)),  # side2 trim at D
                ((20, 10, 2, 0, 2, 10), (0, 2, 0, 8)),  # side2 trim between C and D
                ((20, 10, 2, 0, 10, 10), (0, 2, 0, 0)),  # side2 trim at C
                ((20, 10, 2, 0, 12, 10), (0, 2, 2, 0)),  # side2 trim between B and C
                # side1 trim at C
                ((20, 10, 10, 0, 0, 10), (0, 10, 0, 10)),  # side2 trim at D
                ((20, 10, 10, 0, 2, 10), (0, 10, 0, 8)),  # side2 trim between C and D
                # side1 trim between C and D
                ((20, 10, 12, 0, 0, 10), (0, 10, 0, 8)),  # side2 trim at D
                ((20, 10, 12, 0, 2, 10), (0, 10, 0, 6)),  # side2 trim between C and D
            ):
                assert UrwidImageCanvas._ti_calc_trim(*args) == result

        # A=0, B=5, C=15, D=20
        def test_center_aligned(self):
            for args, result in (
                # side1 trim at A
                ((20, 10, 0, 5, 0, 5), (5, 0, 0, 5)),  # side2 trim at D
                ((20, 10, 0, 5, 2, 5), (5, 0, 0, 3)),  # side2 trim between C and D
                ((20, 10, 0, 5, 5, 5), (5, 0, 0, 0)),  # side2 trim at C
                ((20, 10, 0, 5, 7, 5), (5, 0, 2, 0)),  # side2 trim between B and C
                ((20, 10, 0, 5, 15, 5), (5, 0, 10, 0)),  # side2 trim at B
                ((20, 10, 0, 5, 17, 5), (3, 0, 10, 0)),  # side2 trim between A and B
                # side1 trim between A and B
                ((20, 10, 2, 5, 0, 5), (3, 0, 0, 5)),  # side2 trim at D
                ((20, 10, 2, 5, 2, 5), (3, 0, 0, 3)),  # side2 trim between C and D
                ((20, 10, 2, 5, 5, 5), (3, 0, 0, 0)),  # side2 trim at C
                ((20, 10, 2, 5, 7, 5), (3, 0, 2, 0)),  # side2 trim between B and C
                ((20, 10, 2, 5, 15, 5), (3, 0, 10, 0)),  # side2 trim at B
                ((20, 10, 2, 5, 17, 5), (1, 0, 10, 0)),  # side2 trim between A and B
                # side1 trim at B
                ((20, 10, 5, 5, 0, 5), (0, 0, 0, 5)),  # side2 trim at D
                ((20, 10, 5, 5, 2, 5), (0, 0, 0, 3)),  # side2 trim between C and D
                ((20, 10, 5, 5, 5, 5), (0, 0, 0, 0)),  # side2 trim at C
                ((20, 10, 5, 5, 7, 5), (0, 0, 2, 0)),  # side2 trim between B and C
                # side1 trim between B and C
                ((20, 10, 7, 5, 0, 5), (0, 2, 0, 5)),  # side2 trim at D
                ((20, 10, 7, 5, 2, 5), (0, 2, 0, 3)),  # side2 trim between C and D
                ((20, 10, 7, 5, 5, 5), (0, 2, 0, 0)),  # side2 trim at C
                ((20, 10, 7, 5, 7, 5), (0, 2, 2, 0)),  # side2 trim between B and C
                # side1 trim at C
                ((20, 10, 15, 5, 0, 5), (0, 10, 0, 5)),  # side2 trim at D
                ((20, 10, 15, 5, 2, 5), (0, 10, 0, 3)),  # side2 trim between C and D
                # side1 trim between C and D
                ((20, 10, 17, 5, 0, 5), (0, 10, 0, 3)),  # side2 trim at D
                ((20, 10, 17, 5, 2, 5), (0, 10, 0, 1)),  # side2 trim between C and D
            ):
                assert UrwidImageCanvas._ti_calc_trim(*args) == result

        # A=0, B=10, C=D=20
        def test_side2_aligned(self):
            for args, result in (
                # side1 trim at A
                ((20, 10, 0, 10, 0, 0), (10, 0, 0, 0)),  # side2 trim at C/D
                ((20, 10, 0, 10, 2, 0), (10, 0, 2, 0)),  # side2 trim between B and C
                ((20, 10, 0, 10, 10, 0), (10, 0, 10, 0)),  # side2 trim at B
                ((20, 10, 0, 10, 12, 0), (8, 0, 10, 0)),  # side2 trim between A and B
                # side1 trim between A and B
                ((20, 10, 2, 10, 0, 0), (8, 0, 0, 0)),  # side2 trim at C/D
                ((20, 10, 2, 10, 2, 0), (8, 0, 2, 0)),  # side2 trim between B and C
                ((20, 10, 2, 10, 10, 0), (8, 0, 10, 0)),  # side2 trim at B
                ((20, 10, 2, 10, 12, 0), (6, 0, 10, 0)),  # side2 trim between A and B
                # side1 trim at B
                ((20, 10, 10, 10, 0, 0), (0, 0, 0, 0)),  # side2 trim at C/D
                ((20, 10, 10, 10, 2, 0), (0, 0, 2, 0)),  # side2 trim between B and C
                # side1 trim between B and C
                ((20, 10, 12, 10, 0, 0), (0, 2, 0, 0)),  # side2 trim at C/D
                ((20, 10, 12, 10, 2, 0), (0, 2, 2, 0)),  # side2 trim between B and C
            ):
                assert UrwidImageCanvas._ti_calc_trim(*args) == result

    class TestVertical:
        def _test(self, ImageClass, h_align, v_align, graphics=False):
            render, canv, *_ = get_trim_render_canv(
                ImageClass, (_size[0] - 10, _size[1] - 10), h_align, v_align
            )

            for trim_top, rows in (
                (trim, rows)
                for trim in range(1, _size[1])
                for rows in range(_size[1] - trim, 0, -1)
            ):
                # print(trim_top, rows)
                render_text = [
                    line if graphics else b"".join(line)
                    for line in render[trim_top : trim_top + rows]
                ]
                canv_text = content_to_text(canv.content(0, trim_top, None, rows))

                assert len(render_text) == len(canv_text)
                for render_line, canv_line in zip(render_text, canv_text):
                    render_line += b"\0\0"
                    assert len(render_line) == len(canv_line)
                    assert render_line == canv_line

        def test_top_left_aligned_Text(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "<", "^")

        def test_top_center_aligned_Text(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "|", "^")

        def test_top_right_aligned_Text(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, ">", "^")

        def test_middle_left_aligned_Text(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "<", "-")

        def test_middle_center_aligned_Text(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "|", "-")

        def test_middle_right_aligned_Text(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, ">", "-")

        def test_bottom_left_aligned_Text(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "<", "_")

        def test_bottom_center_aligned_Text(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "|", "_")

        def test_bottom_right_aligned_Text(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, ">", "_")

        def test_top_left_aligned_Graphics(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "<", "^", graphics=True)

        def test_top_center_aligned_Graphics(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "|", "^", graphics=True)

        def test_top_right_aligned_Graphics(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, ">", "^", graphics=True)

        def test_middle_left_aligned_Graphics(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "<", "-", graphics=True)

        def test_middle_center_aligned_Graphics(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "|", "-", graphics=True)

        def test_middle_right_aligned_Graphics(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, ">", "-", graphics=True)

        def test_bottom_left_aligned_Graphics(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "<", "_", graphics=True)

        def test_bottom_center_aligned_Graphics(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "|", "_", graphics=True)

        def test_bottom_right_aligned_Graphics(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, ">", "_", graphics=True)

    class TestHorizontalText:
        def _test(self, ImageClass, h_align, v_align):
            (
                render,
                canv,
                pad_top,
                pad_bottom,
                pad_left,
                pad_right,
            ) = get_trim_render_canv(
                ImageClass, (_size[0] - 10, _size[1] - 10), h_align, v_align
            )

            for trim_left, cols in get_hori_trim_args():
                # print(trim_left, cols)
                trim_right = _size[0] - trim_left - cols
                render_text = [
                    b"".join(line[trim_left : trim_left + cols]) for line in render
                ]
                canv_text = content_to_text(canv.content(trim_left, 0, cols))
                prefix = ctlseqs.SGR_NORMAL_b * (
                    pad_left < trim_left < _size[0] - pad_right
                )
                suffix = ctlseqs.SGR_NORMAL_b * (
                    _size[0] - pad_left > trim_right > pad_right
                )

                assert len(render_text) == len(canv_text)

                row = 0
                for render_line, canv_line in zip(render_text, canv_text):
                    # Exclude padding lines
                    if pad_top <= row < _size[1] - pad_bottom:
                        render_line = prefix + render_line + suffix
                    render_line += b"\0\0"
                    row += 1

                    assert len(render_line) == len(canv_line)
                    assert render_line == canv_line

        def test_top_left_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "<", "^")

        def test_top_center_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "|", "^")

        def test_top_right_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, ">", "^")

        def test_middle_left_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "<", "-")

        def test_middle_center_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "|", "-")

        def test_middle_right_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, ">", "-")

        def test_bottom_left_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "<", "_")

        def test_bottom_center_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "|", "_")

        def test_bottom_right_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, ">", "_")

    class TestHorizontalGraphics:
        def _test(self, ImageClass, h_align, v_align):
            render, canv = get_trim_render_canv(
                ImageClass, (_size[0] - 10, _size[1] - 10), h_align, v_align
            )

            for trim_left, cols in get_hori_trim_args():
                # print(trim_left, cols)
                render_text = [b" " * cols] * _size[1]
                canv_text = content_to_text(canv.content(trim_left, 0, cols))

                assert len(render_text) == len(canv_text)
                for render_line, canv_line in zip(render_text, canv_text):
                    assert len(render_line) == len(canv_line)
                    assert render_line == canv_line

        def test_top_left_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "<", "^")

        def test_top_center_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "|", "^")

        def test_top_right_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, ">", "^")

        def test_middle_left_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "<", "-")

        def test_middle_center_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "|", "-")

        def test_middle_right_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, ">", "-")

        def test_bottom_left_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "<", "_")

        def test_bottom_center_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "|", "_")

        def test_bottom_right_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, ">", "_")

    class TestBothText:
        def _test(self, ImageClass, h_align, v_align):
            (
                render,
                canv,
                pad_top,
                pad_bottom,
                pad_left,
                pad_right,
            ) = get_trim_render_canv(
                ImageClass, (_size[0] - 10, _size[1] - 10), h_align, v_align
            )

            for trim_left, trim_top, cols, rows in get_all_trim_args():
                # print(trim_left, trim_top, cols, rows)
                trim_right = _size[0] - trim_left - cols
                render_text = [
                    b"".join(line[trim_left : trim_left + cols])
                    for line in render[trim_top : trim_top + rows]
                ]
                canv_text = content_to_text(
                    canv.content(trim_left, trim_top, cols, rows)
                )
                prefix = ctlseqs.SGR_NORMAL_b * (
                    pad_left < trim_left < _size[0] - pad_right
                )
                suffix = ctlseqs.SGR_NORMAL_b * (
                    _size[0] - pad_left > trim_right > pad_right
                )

                assert len(render_text) == len(canv_text)

                row = trim_top
                for render_line, canv_line in zip(render_text, canv_text):
                    # Exclude padding lines
                    if pad_top <= row < _size[1] - pad_bottom:
                        render_line = prefix + render_line + suffix
                    render_line += b"\0\0"
                    row += 1

                    assert len(render_line) == len(canv_line)
                    assert render_line == canv_line

        def test_top_left_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "<", "^")

        def test_top_center_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "|", "^")

        def test_top_right_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, ">", "^")

        def test_middle_left_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "<", "-")

        def test_middle_center_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "|", "-")

        def test_middle_right_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, ">", "-")

        def test_bottom_left_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "<", "_")

        def test_bottom_center_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, "|", "_")

        def test_bottom_right_aligned(self):
            for cls in TextImage.__subclasses__():
                self._test(cls, ">", "_")

    class TestBothGraphics:
        def _test(self, ImageClass, h_align, v_align):
            render, canv = get_trim_render_canv(
                ImageClass, (_size[0] - 10, _size[1] - 10), h_align, v_align
            )

            for trim_left, trim_top, cols, rows in get_all_trim_args():
                # print(trim_left, trim_top, cols, rows)
                render_text = [b" " * cols] * rows
                canv_text = content_to_text(
                    canv.content(trim_left, trim_top, cols, rows)
                )

                assert len(render_text) == len(canv_text)
                for render_line, canv_line in zip(render_text, canv_text):
                    assert len(render_line) == len(canv_line)
                    assert render_line == canv_line

        def test_top_left_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "<", "^")

        def test_top_center_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "|", "^")

        def test_top_right_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, ">", "^")

        def test_middle_left_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "<", "-")

        def test_middle_center_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "|", "-")

        def test_middle_right_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, ">", "-")

        def test_bottom_left_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "<", "_")

        def test_bottom_center_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, "|", "_")

        def test_bottom_right_aligned(self):
            for cls in GraphicsImage.__subclasses__():
                self._test(cls, ">", "_")

    def test_horizontal_trim_first_color_ColoredText(self):
        for ImageClass in TextImage.__subclasses__():
            image = ImageClass.from_file("tests/images/trans.png")
            image._size = _size
            render = image._renderer(image._render_image, "#102030", split_cells=True)
            image_w = UrwidImage(image, "#102030")
            canv = UrwidImageCanvas(render, _size, _size)
            canv._widget_info = (image_w, _size, False)

            for trim_left in range(1, _size[0]):
                text = content_to_text(
                    canv.content(trim_left, cols=_size[0] - trim_left)
                )
                for line in text:
                    assert line.startswith(ctlseqs.SGR_BG_RGB_b % (16, 32, 48) + b" ")


class TestScreen:
    class TestStartStop:
        def test_supported(self):
            buf = io.StringIO()
            screen = UrwidImageScreen(sys.__stdin__, buf)

            screen.start()
            start_output = buf.getvalue()
            buf.seek(0)
            buf.truncate()
            screen.stop()
            stop_output = buf.getvalue()

            assert start_output.endswith(ctlseqs.KITTY_DELETE_ALL)
            assert stop_output.startswith(ctlseqs.KITTY_DELETE_ALL)

        def test_not_supported(self):
            buf = io.StringIO()
            screen = UrwidImageScreen(sys.__stdin__, buf)

            KittyImage._supported = False
            try:
                screen.start()
                start_output = buf.getvalue()
                buf.seek(0)
                buf.truncate()
                screen.stop()
                stop_output = buf.getvalue()

                assert ctlseqs.KITTY_DELETE_ALL not in start_output
                assert ctlseqs.KITTY_DELETE_ALL not in stop_output
            finally:
                KittyImage._supported = True

    def test_synced_output(self):
        widget = urwid.SolidFill("x")
        buf = io.StringIO()
        screen = UrwidImageScreen(sys.__stdin__, buf)
        screen.start()

        buf.seek(0)
        buf.truncate()
        screen.draw_screen(_size, widget.render(_size))
        output = buf.getvalue()

        assert output.startswith(ctlseqs.BEGIN_SYNCED_UPDATE)
        assert output.endswith(ctlseqs.END_SYNCED_UPDATE)


buf = io.StringIO()
screen = UrwidImageScreen(sys.__stdin__, buf)
screen.start()


@contextmanager
def setup_clear_buffers():
    from term_image.widget import urwid

    tty_buf = io.BytesIO()
    write_tty = urwid.write_tty
    urwid.write_tty = tty_buf.write
    buf.seek(0)
    buf.truncate()
    try:
        yield buf, tty_buf
    finally:
        buf.seek(0)
        buf.truncate()
        tty_buf.close()
        urwid.write_tty = write_tty


class TestScreenClearImages:
    class TestNowFalse:
        def test_single(self):
            image_ws = [UrwidImage(kitty_image) for _ in range(4)]
            for image_w in image_ws:
                with setup_clear_buffers() as (buf, tty_buf):
                    screen.clear_images(image_w)
                    assert (
                        buf.getvalue()
                        == ctlseqs.KITTY_DELETE_Z_INDEX % image_w._ti_z_index
                    )
                    assert tty_buf.getvalue() == b""

        def test_multiple(self):
            image_ws = [UrwidImage(kitty_image) for _ in range(4)]
            with setup_clear_buffers() as (buf, tty_buf):
                screen.clear_images(*image_ws)
                assert buf.getvalue() == "".join(
                    ctlseqs.KITTY_DELETE_Z_INDEX % image_w._ti_z_index
                    for image_w in image_ws
                )
                assert tty_buf.getvalue() == b""

        def test_multiple_not_kitty(self):
            image_ws = [UrwidImage(trans) for _ in range(4)]
            with setup_clear_buffers() as (buf, tty_buf):
                screen.clear_images(*image_ws)
                assert buf.getvalue() == ""
                assert tty_buf.getvalue() == b""

        def test_all(self):
            with setup_clear_buffers() as (buf, tty_buf):
                screen.clear_images()
                assert buf.getvalue() == ctlseqs.KITTY_DELETE_ALL
                assert tty_buf.getvalue() == b""

    class TestNowTrue:
        def test_single(self):
            image_ws = [UrwidImage(kitty_image) for _ in range(4)]
            for image_w in image_ws:
                with setup_clear_buffers() as (buf, tty_buf):
                    screen.clear_images(image_w, now=True)
                    assert buf.getvalue() == ""
                    assert (
                        tty_buf.getvalue()
                        == ctlseqs.KITTY_DELETE_Z_INDEX_b % image_w._ti_z_index
                    )

        def test_multiple(self):
            image_ws = [UrwidImage(kitty_image) for _ in range(4)]
            with setup_clear_buffers() as (buf, tty_buf):
                screen.clear_images(*image_ws, now=True)
                assert buf.getvalue() == ""
                assert tty_buf.getvalue() == b"".join(
                    ctlseqs.KITTY_DELETE_Z_INDEX_b % image_w._ti_z_index
                    for image_w in image_ws
                )

        def test_multiple_not_kitty(self):
            image_ws = [UrwidImage(trans) for _ in range(4)]
            with setup_clear_buffers() as (buf, tty_buf):
                screen.clear_images(*image_ws, now=True)
                assert buf.getvalue() == ""
                assert tty_buf.getvalue() == b""

        def test_all(self):
            with setup_clear_buffers() as (buf, tty_buf):
                screen.clear_images(now=True)
                assert buf.getvalue() == ""
                assert tty_buf.getvalue() == ctlseqs.KITTY_DELETE_ALL_b

    def test_not_supported(self):
        image_ws = [UrwidImage(kitty_image) for _ in range(4)]
        try:
            KittyImage._supported = False
            with setup_clear_buffers() as (buf, tty_buf):
                screen.clear_images(*image_ws)
                assert buf.getvalue() == ""
                assert tty_buf.getvalue() == b""

                screen.clear_images(*image_ws, now=True)
                assert buf.getvalue() == ""
                assert tty_buf.getvalue() == b""
        finally:
            KittyImage._supported = True

    def test_not_supported_all(self):
        try:
            KittyImage._supported = False
            with setup_clear_buffers() as (buf, tty_buf):
                screen.clear_images()
                assert buf.getvalue() == ""
                assert tty_buf.getvalue() == b""

                screen.clear_images(now=True)
                assert buf.getvalue() == ""
                assert tty_buf.getvalue() == b""
        finally:
            KittyImage._supported = True

    def test_disguise_state(self):
        image_w = UrwidImage(kitty_image)
        assert image_w._ti_disguise_state == 0

        with setup_clear_buffers():
            for value in (1, 2, 0, 1, 2, 0, 1):
                screen.clear_images(image_w)
                assert image_w._ti_disguise_state == value

    def test_disguise_state_all(self):
        try:
            UrwidImageCanvas._ti_disguise_state = 0
            with setup_clear_buffers():
                for value in (1, 2, 0, 1, 2, 0, 1):
                    screen.clear_images()
                    assert UrwidImageCanvas._ti_disguise_state == value
        finally:
            UrwidImageCanvas._ti_disguise_state = 0


block_image_w = UrwidImage(BlockImage(python_img))
kitty_image_w = UrwidImage(KittyImage(python_img))
iterm2_image_w = UrwidImage(ITerm2Image(python_img))

divider = urwid.SolidFill("\u2500")
kitty_list_box = urwid.ListBox(
    [urwid.Pile([("pack", kitty_image_w), (1, divider)] * 3)]
)
iterm2_list_box = urwid.ListBox(
    [urwid.Pile([("pack", iterm2_image_w), (1, divider)] * 3)]
)
bottom_w = urwid.Columns(
    [urwid.LineBox(kitty_list_box), urwid.LineBox(iterm2_list_box)]
)
top_w = urwid.LineBox(urwid.SolidFill("x"))
widget = urwid.Overlay(
    top_w,
    bottom_w,
    "center",
    10,
    "top",
    5,
)


class TestScreenAutoClearImages:
    def test_image_cviews(self):
        assert screen._ti_image_cviews == frozenset()
        screen.draw_screen(_size, widget.render(_size))
        assert isinstance(screen._ti_image_cviews, frozenset)

    def test_move_top_widget(self):
        _terminal_name_version = get_terminal_name_version()
        try:
            set_terminal_name_version("konsole")
            # Setup
            buf.seek(0)
            buf.truncate()
            widget.top_w = top_w
            kitty_list_box.shift_focus(_size, 0)
            iterm2_list_box.shift_focus(_size, 0)

            widget.top = 2
            widget._invalidate()
            canv = widget.render(_size)
            # print(b"\n".join(canv.text).decode())
            screen.draw_screen(_size, canv)

            assert len(screen._ti_image_cviews) == 6
            assert {(row, col) for _, row, col, *_ in screen._ti_image_cviews} == {
                (2, 2),
                (2, 17),
                (3, 2),
                (3, 21),
                (9, 2),
                (9, 17),
            }

            prev_image_cviews = screen._ti_image_cviews
            widget.top += 1
            widget._invalidate()
            canv = widget.render(_size)
            # print(b"\n".join(canv.text).decode())
            screen.draw_screen(_size, canv)

            assert len(screen._ti_image_cviews) == 6
            assert prev_image_cviews != screen._ti_image_cviews
            # The two lower cviews should be unchanged
            assert len(prev_image_cviews & screen._ti_image_cviews) == 2
            assert {(row, col) for _, row, col, *_ in screen._ti_image_cviews} == {
                (2, 2),
                (2, 17),
                (4, 2),
                (4, 21),
                (9, 2),
                (9, 17),
            }

            prev_image_cviews = screen._ti_image_cviews
            widget.top = 5
            widget._invalidate()
            canv = widget.render(_size)
            # print(b"\n".join(canv.text).decode())
            screen.draw_screen(_size, canv)

            assert len(screen._ti_image_cviews) == 8
            assert prev_image_cviews != screen._ti_image_cviews
            # All cviews should be changed
            assert len(prev_image_cviews & screen._ti_image_cviews) == 0
            assert {(row, col) for _, row, col, *_ in screen._ti_image_cviews} == {
                (2, 2),
                (2, 17),
                (6, 2),
                (6, 21),
                (9, 2),
                (9, 21),
                (11, 2),
                (11, 17),
            }

            prev_image_cviews = screen._ti_image_cviews
            widget.top = 8
            widget._invalidate()
            canv = widget.render(_size)
            # print(b"\n".join(canv.text).decode())
            screen.draw_screen(_size, canv)

            assert len(screen._ti_image_cviews) == 6
            assert prev_image_cviews != screen._ti_image_cviews
            # All cviews should be changed
            assert len(prev_image_cviews & screen._ti_image_cviews) == 0
            assert {(row, col) for _, row, col, *_ in screen._ti_image_cviews} == {
                (2, 2),
                (2, 17),
                (9, 2),
                (9, 21),
                (14, 2),
                (14, 17),
            }

        finally:
            set_terminal_name_version(*_terminal_name_version)

    # FIXME: Figure out why the listbox insets are not scrolling the listbox by the
    # expected number of rows
    def test_scroll_listboxes(self):
        _terminal_name_version = get_terminal_name_version()
        try:
            set_terminal_name_version("konsole")
            # Setup
            buf.seek(0)
            buf.truncate()
            widget.top_w = top_w
            widget.top = 5
            kitty_list_box.shift_focus(_size, 0)
            iterm2_list_box.shift_focus(_size, 0)
            widget._invalidate()
            screen.draw_screen(_size, widget.render(_size))

            prev_image_cviews = screen._ti_image_cviews
            kitty_list_box.shift_focus(_size, -3)
            canv = widget.render(_size)
            # print(b"\n".join(canv.text).decode())
            screen.draw_screen(_size, canv)

            assert len(screen._ti_image_cviews) == 8
            assert prev_image_cviews != screen._ti_image_cviews
            # All images (4 cviews) on the right should remain unchanged
            assert len(prev_image_cviews & screen._ti_image_cviews) == 4
            assert {(row, col) for _, row, col, *_ in screen._ti_image_cviews} == {
                (2, 2),
                (2, 17),
                (6, 2),
                (6, 21),
                (8, 2),
                (9, 21),
                (11, 2),
                (11, 17),
            }
            assert {
                (row, col)
                for _, row, col, *_ in prev_image_cviews & screen._ti_image_cviews
            } == {
                (2, 17),
                (6, 21),
                (9, 21),
                (11, 17),
            }

            prev_image_cviews = screen._ti_image_cviews
            iterm2_list_box.shift_focus(_size, -9)
            canv = widget.render(_size)
            # print(b"\n".join(canv.text).decode())
            screen.draw_screen(_size, canv)

            assert len(screen._ti_image_cviews) == 8
            assert prev_image_cviews != screen._ti_image_cviews
            # All images (4 cviews) on the left should remain unchanged
            assert len(prev_image_cviews & screen._ti_image_cviews) == 4
            assert {(row, col) for _, row, col, *_ in screen._ti_image_cviews} == {
                (2, 2),
                (2, 17),
                (6, 2),
                (6, 21),
                (8, 2),
                (11, 2),
                (11, 17),
                (13, 17),
            }
            assert {
                (row, col)
                for _, row, col, *_ in prev_image_cviews & screen._ti_image_cviews
            } == {
                (2, 2),
                (6, 2),
                (8, 2),
                (11, 2),
            }

        finally:
            set_terminal_name_version(*_terminal_name_version)

    def test_change_top_widget(self):
        _terminal_name_version = get_terminal_name_version()
        try:
            set_terminal_name_version("konsole")
            # Setup
            buf.seek(0)
            buf.truncate()
            widget.top_w = top_w
            widget.top = 5
            kitty_list_box.shift_focus(_size, 0)
            iterm2_list_box.shift_focus(_size, 0)
            screen.draw_screen(_size, widget.render(_size))

            # kitty image in the top widget
            prev_image_cviews = screen._ti_image_cviews
            widget.top_w = kitty_image_w
            widget._invalidate()
            canv = widget.render(_size)
            # print(b"\n".join(canv.text).decode())
            screen.draw_screen(_size, canv)

            assert len(screen._ti_image_cviews) == 9
            assert prev_image_cviews != screen._ti_image_cviews
            assert len(screen._ti_image_cviews - prev_image_cviews) == 1
            assert (*(screen._ti_image_cviews - prev_image_cviews),)[0][1:3] == (6, 11)

            # block image in the top widget
            prev_image_cviews = screen._ti_image_cviews
            widget.top_w = block_image_w
            widget._invalidate()
            canv = widget.render(_size)
            # print(b"\n".join(canv.text).decode())
            screen.draw_screen(_size, canv)

            assert len(screen._ti_image_cviews) == 8
            assert prev_image_cviews != screen._ti_image_cviews
            assert len(prev_image_cviews - screen._ti_image_cviews) == 1
            assert (*(prev_image_cviews - screen._ti_image_cviews),)[0][1:3] == (6, 11)

            # iterm2 image in the top widget
            prev_image_cviews = screen._ti_image_cviews
            widget.top_w = iterm2_image_w
            widget._invalidate()
            canv = widget.render(_size)
            # print(b"\n".join(canv.text).decode())
            screen.draw_screen(_size, canv)

            assert len(screen._ti_image_cviews) == 9
            assert prev_image_cviews != screen._ti_image_cviews
            assert len(screen._ti_image_cviews - prev_image_cviews) == 1
            assert (*(screen._ti_image_cviews - prev_image_cviews),)[0][1:3] == (6, 11)

        finally:
            set_terminal_name_version(*_terminal_name_version)

    def test_iterm2_not_on_konsole(self):
        _terminal_name_version = get_terminal_name_version()
        try:
            set_terminal_name_version("wezterm")
            # Setup
            buf.seek(0)
            buf.truncate()
            widget.top_w = top_w
            widget.top = 5
            kitty_list_box.shift_focus(_size, 0)
            iterm2_list_box.shift_focus(_size, 0)

            widget._invalidate()
            canv = widget.render(_size)  # noqa: F841
            # print(b"\n".join(canv.text).decode())
            screen.draw_screen(_size, canv)

            # The images on the right are of iterm2 style but not on konsole
            assert len(screen._ti_image_cviews) == 4
            assert {(row, col) for _, row, col, *_ in screen._ti_image_cviews} == {
                (2, 2),
                (6, 2),
                (9, 2),
                (11, 2),
            }

        finally:
            set_terminal_name_version(*_terminal_name_version)



from operator import truediv
from random import randint, random

import pytest

from term_image import AutoCellRatio, get_cell_ratio, set_cell_ratio
from term_image.exceptions import TermImageError

from . import reset_cell_size_ratio, set_cell_size


class TestCellRatio:
    @reset_cell_size_ratio()
    def test_args(self):
        for value in (0, "1", ()):
            with pytest.raises(TypeError):
                set_cell_ratio(value)
        for value in (0.0, -0.1, -1.0):
            with pytest.raises(ValueError):
                set_cell_ratio(value)

        set_cell_size(None)
        for value in AutoCellRatio:
            with pytest.raises(TermImageError):
                set_cell_ratio(value)

    @reset_cell_size_ratio()
    def test_fixed(self):
        for value in (0.01, 0.1, 0.9, 0.99, 1.0, 2.0):
            set_cell_ratio(value)
            assert get_cell_ratio() == value

        for _ in range(20):
            value = random()
            set_cell_ratio(value)
            assert get_cell_ratio() == value

    @reset_cell_size_ratio()
    def test_fixed_auto(self):
        set_cell_size((4, 9))
        set_cell_ratio(AutoCellRatio.FIXED)
        assert get_cell_ratio() == 4 / 9 == get_cell_ratio()

        for _ in range(20):
            cell_size = (randint(1, 20), randint(1, 20))
            set_cell_size(cell_size)
            set_cell_ratio(AutoCellRatio.FIXED)
            assert get_cell_ratio() == truediv(*cell_size) == get_cell_ratio()
            set_cell_size((0, 1))
            assert get_cell_ratio() == truediv(*cell_size)

    @reset_cell_size_ratio()
    def test_dynamic_auto(self):
        set_cell_size((4, 9))
        set_cell_ratio(AutoCellRatio.DYNAMIC)
        assert get_cell_ratio() == 4 / 9 == get_cell_ratio()

        for _ in range(20):
            cell_size = (randint(1, 20), randint(1, 20))
            set_cell_size(cell_size)
            assert get_cell_ratio() == truediv(*cell_size) == get_cell_ratio()



"""Render-style-dependent (though shared, not specific) tests"""

import atexit
from operator import gt, lt, mul
from types import SimpleNamespace

import pytest
from PIL import Image

from term_image import set_cell_ratio
from term_image.ctlseqs import SGR_BG_RGB, SGR_NORMAL
from term_image.exceptions import StyleError
from term_image.image.common import _ALPHA_THRESHOLD, GraphicsImage, Size, TextImage
from term_image.utils import get_fg_bg_colors, get_terminal_size

from .. import (
    reset_cell_size_ratio,
    set_cell_size,
    set_fg_bg_colors,
    toggle_is_on_kitty,
)

columns, lines = get_terminal_size()

# For square images
dummy = SimpleNamespace()
dummy._original_size = (1, 1)
_width = _height = _width_px = _height_px = None  # Set by `setup_common()`

_size = 20
ImageClass = None  # Set by `setup_common()`
python_img = Image.open("tests/images/python.png")


@atexit.register
def close_imgs():
    python_img.close()


def setup_common(ImageClass):
    globals().update(locals())  # width_height() requires ImageClass

    set_cell_size((9, 18))
    set_cell_ratio(0.5)

    if ImageClass._pixels_cols(cols=columns) < ImageClass._pixels_lines(
        lines=lines - 2
    ):
        _width = columns
        _height = width_height(dummy, w=columns)
    else:
        _height = lines - 2
        _width = width_height(dummy, h=lines - 2)
    _width_px = ImageClass._pixels_cols(cols=_width)
    _height_px = ImageClass._pixels_lines(lines=_height)

    globals().update(locals())


def proportional(image):
    ori_width, ori_height = image.original_size
    # The converted width and height in pixels might not be exactly proportional
    # as they have been previously rounded to meet cell boundaries but the
    # difference must be less than the number of pixels for one line
    if _width_px < _height_px:
        # Height was adjusted
        return (
            abs(
                image._pixels_cols(cols=image.width) / ori_width
                # Was adjusted by multiplying, so we divide
                - (image._pixels_lines(lines=image.height) / image._pixel_ratio)
                / ori_height
            )
            < image._pixels_lines(lines=1) / ori_height
        )
    else:
        # Width was adjusted
        return (
            abs(
                # Was adjusted by dividing, so we multiply
                (image._pixels_cols(cols=image.width) * image._pixel_ratio) / ori_width
                - image._pixels_lines(lines=image.height) / ori_height
            )
            < image._pixels_lines(lines=1) / ori_height
        )


def width_height(image, *, w=None, h=None):
    return (
        ImageClass._pixels_lines(
            pixels=round(
                ImageClass._width_height_px(
                    image,
                    w=ImageClass._pixels_cols(cols=w),
                )
            )
        )
        if w is not None
        else ImageClass._pixels_cols(
            pixels=round(
                ImageClass._width_height_px(
                    image,
                    h=ImageClass._pixels_lines(lines=h),
                )
            )
        )
    )


def get_actual_render_size(image):
    render_size = image._get_render_size()

    if image._render_method != "whole":
        return render_size

    _, r_height = image.rendered_size
    width, height = (
        render_size
        if mul(*render_size) < mul(*image._original_size)
        else image._original_size
    )
    if image._render_method != "whole":
        extra = height % (r_height or 1)
        if extra:
            height = height - extra + r_height

    return width, height


def test_instantiation_Text():
    original = ImageClass._supported
    try:
        ImageClass._supported = True
        assert isinstance(ImageClass(python_img), TextImage)
        ImageClass._supported = False
        assert isinstance(ImageClass(python_img), TextImage)
    finally:
        ImageClass._supported = original


def test_instantiation_Graphics():
    original = ImageClass._supported
    try:
        ImageClass._supported = True
        assert isinstance(ImageClass(python_img), GraphicsImage)
        ImageClass._supported = False
        with pytest.raises(StyleError):
            ImageClass(python_img)
    finally:
        ImageClass._supported = original


def test_str_All():
    image = ImageClass(python_img, width=_size)
    assert str(image) == image._render_image(python_img, _ALPHA_THRESHOLD)


def test_format_All():
    image = ImageClass(python_img)
    image.set_size()
    image.width //= 2  # To ensure there's padding
    assert format(image) == image._format_render(str(image), *image._check_formatting())


def test_is_supported_All():
    assert isinstance(ImageClass.is_supported(), bool)


def test_forced_support_Graphics():
    original = ImageClass._supported
    try:
        ImageClass._supported = False
        with pytest.raises(StyleError):
            ImageClass(python_img)

        ImageClass.forced_support = True
        assert isinstance(ImageClass(python_img), GraphicsImage)

        ImageClass.forced_support = False
        with pytest.raises(StyleError):
            ImageClass(python_img)
    finally:
        ImageClass._supported = original


def test_set_render_method_All():
    for value in (2, 2.0, ()):
        with pytest.raises(TypeError):
            ImageClass.set_render_method(value)

    default = ImageClass._default_render_method if ImageClass._render_methods else None

    with pytest.raises(ValueError):
        ImageClass.set_render_method("")

    assert ImageClass._render_method == default
    assert ImageClass.set_render_method() is None
    for method in ImageClass._render_methods:
        assert ImageClass.set_render_method(method) is None

    assert ImageClass.set_render_method(None) is None
    assert ImageClass._render_method == default
    if default is None:
        assert "_render_method" not in vars(ImageClass)

    image = ImageClass(python_img)

    with pytest.raises(ValueError):
        image.set_render_method("")

    assert image._render_method == default
    assert image.set_render_method() is None
    for method in image._render_methods:
        assert image.set_render_method(method) is None

    assert image.set_render_method(None) is None
    assert image._render_method == default
    if default is None:
        assert "_render_method" not in vars(image)


# Size-setting is taken as style-dependent because the major underlying interface is
# style-specific
class TestSetSize_All:
    def test_setup(self):
        type(self).image = ImageClass(python_img)  # Square
        # Horizontally-oriented
        type(self).h_image = ImageClass.from_file("tests/images/hori.jpg")
        # Vertically-oriented
        type(self).v_image = ImageClass.from_file("tests/images/vert.jpg")

    def test_auto(self):
        frame_width_px = ImageClass._pixels_cols(cols=columns)
        frame_height_px = ImageClass._pixels_lines(lines=lines - 2)

        _original_size = self.image.original_size
        try:
            # Equal to frame size
            self.image._original_size = (frame_width_px, frame_height_px)
            self.image.set_size(Size.AUTO)
            assert self.image._valid_size(Size.ORIGINAL) == self.image.size

            # Horizontally larger than frame size
            self.image._original_size = (frame_width_px + 20, frame_height_px - 20)
            self.image.set_size(Size.AUTO)
            assert (
                self.image._valid_size(Size.ORIGINAL)
                != self.image.size
                == self.image._valid_size(Size.FIT)
            )

            # Vertically larger than frame size
            self.image._original_size = (frame_width_px - 20, frame_height_px + 20)
            self.image.set_size(Size.AUTO)
            assert (
                self.image._valid_size(Size.ORIGINAL)
                != self.image.size
                == self.image._valid_size(Size.FIT)
            )

            # Horizontally and vertically larger than frame size
            self.image._original_size = (frame_width_px + 20, frame_height_px + 20)
            self.image.set_size(Size.AUTO)
            assert (
                self.image._valid_size(Size.ORIGINAL)
                != self.image.size
                == self.image._valid_size(Size.FIT)
            )

            # Smaller than frame size
            self.image._original_size = (frame_width_px - 20, frame_height_px - 20)
            self.image.set_size(Size.AUTO)
            assert (
                self.image._valid_size(Size.ORIGINAL)
                == self.image.size
                != self.image._valid_size(Size.FIT)
            )
        finally:
            self.image._original_size = _original_size

    def test_fit_none_default(self):
        self.image.set_size()
        size = self.image.size

        self.image.set_size(width=Size.FIT)
        assert self.image.size == size

        self.image.set_size(height=Size.FIT)
        assert self.image.size == size

        self.image.set_size(width=None)
        assert self.image.size == size

        self.image.set_size(height=None)
        assert self.image.size == size

    # a PASS is valid only if the previous test passed
    def test_fit(self):
        self.image.set_size()
        assert self.image.size == (_width, _height) == self.image._size
        assert proportional(self.image)

        self.h_image.set_size()
        assert gt(
            self.h_image._pixels_cols(cols=self.h_image.width),
            self.h_image._pixels_lines(lines=self.h_image.height),
        )
        assert proportional(self.h_image)

        self.v_image.set_size()
        assert lt(
            self.v_image._pixels_cols(cols=self.v_image.width),
            self.v_image._pixels_lines(lines=self.v_image.height),
        )
        assert proportional(self.v_image)

    def test_fit_to_width_width(self):
        self.image.set_size(width=Size.FIT_TO_WIDTH)
        assert self.image.width == columns
        assert proportional(self.image)

        self.h_image.set_size(width=Size.FIT_TO_WIDTH)
        assert self.h_image.width == columns
        assert proportional(self.h_image)

        self.v_image.set_size(width=Size.FIT_TO_WIDTH)
        assert self.v_image.width == columns
        assert proportional(self.v_image)

    def test_fit_to_width_height(self):
        self.image.set_size(height=Size.FIT_TO_WIDTH)
        assert self.image.width == columns
        assert proportional(self.image)

        self.h_image.set_size(height=Size.FIT_TO_WIDTH)
        assert self.h_image.width == columns
        assert proportional(self.h_image)

        self.v_image.set_size(height=Size.FIT_TO_WIDTH)
        assert self.v_image.width == columns
        assert proportional(self.v_image)

    def test_int_width(self):
        self.image.set_size(width=_size)
        assert self.image.width == _size
        assert proportional(self.image)

        self.h_image.set_size(width=_size)
        assert self.h_image.width == _size
        assert proportional(self.h_image)

        self.v_image.set_size(width=_size)
        assert self.v_image.width == _size
        assert proportional(self.v_image)

    def test_int_height(self):
        self.image.set_size(height=_size)
        assert self.image.height == _size
        assert proportional(self.image)

        self.h_image.set_size(height=_size)
        assert self.h_image.height == _size
        assert proportional(self.h_image)

        self.v_image.set_size(height=_size)
        assert self.v_image.height == _size
        assert proportional(self.v_image)

    def test_original_width(self):
        ori_width, ori_height = self.image.original_size
        self.image.set_size(width=Size.ORIGINAL)
        assert self.image.width == ImageClass._pixels_cols(pixels=ori_width)
        assert proportional(self.image)

        ori_width, ori_height = self.h_image.original_size
        self.h_image.set_size(width=Size.ORIGINAL)
        assert self.h_image.width == ImageClass._pixels_cols(pixels=ori_width)
        assert proportional(self.h_image)

        ori_width, ori_height = self.v_image.original_size
        self.v_image.set_size(width=Size.ORIGINAL)
        assert self.v_image.width == ImageClass._pixels_cols(pixels=ori_width)
        assert proportional(self.v_image)

    def test_original_height(self):
        ori_width, ori_height = self.image.original_size
        self.image.set_size(height=Size.ORIGINAL)
        assert self.image.height == ImageClass._pixels_lines(pixels=ori_height)
        assert proportional(self.image)

        ori_width, ori_height = self.h_image.original_size
        self.h_image.set_size(height=Size.ORIGINAL)
        assert self.h_image.height == ImageClass._pixels_lines(pixels=ori_height)
        assert proportional(self.h_image)

        ori_width, ori_height = self.v_image.original_size
        self.v_image.set_size(height=Size.ORIGINAL)
        assert self.v_image.height == ImageClass._pixels_lines(pixels=ori_height)
        assert proportional(self.v_image)

    def test_can_exceed_terminal_size(self):
        self.image.set_size(width=columns + 1)
        assert self.image.width == columns + 1
        assert proportional(self.image)

        self.image.set_size(height=lines + 1)
        assert self.image.height == lines + 1
        assert proportional(self.image)


class TestCellRatio_Text:
    def test_setup(self):
        type(self).image = ImageClass(python_img)  # Square

    @reset_cell_size_ratio()
    def test_fixed_width_cell_ratio_adjustment(self):
        for ratio in (0.01, 0.1, 0.25, 0.4, 0.45, 0.55, 0.6, 0.75, 0.9, 0.99, 1.0):
            set_cell_ratio(ratio)
            self.image.set_size(width=_size)
            assert self.image.width == _size
            assert proportional(self.image)

    @reset_cell_size_ratio()
    def test_fixed_height_cell_ratio_adjustment(self):
        for ratio in (0.01, 0.1, 0.25, 0.4, 0.45, 0.55, 0.6, 0.75, 0.9, 0.99, 1.0):
            set_cell_ratio(ratio)
            self.image.set_size(height=_size)
            assert self.image.height == _size
            assert proportional(self.image)

    @reset_cell_size_ratio()
    def test_auto_size_cell_ratio_adjustment(self):
        for ratio in (0.01, 0.1, 0.25, 0.4, 0.45, 0.55, 0.6, 0.75, 0.9, 0.99, 1.0):
            set_cell_ratio(ratio)
            self.image.set_size()
            assert proportional(self.image)


class TestCellRatio_Graphics:
    def test_setup(self):
        type(self).image = ImageClass(python_img)  # Square

    @reset_cell_size_ratio()
    def test_cell_ratio_adjustment(self):
        self.image.set_size()
        size = self.image.size
        for ratio in (0.01, 0.1, 0.25, 0.4, 0.45, 0.55, 0.6, 0.75, 0.9, 0.99, 1.0):
            set_cell_ratio(ratio)
            self.image.set_size()
            assert self.image.size == size


class TestRender_Text:
    def test_setup(self):
        type(self).trans = ImageClass.from_file("tests/images/trans.png", height=_size)

    def render_image(self, alpha, *, split_cells=False):
        return self.trans._renderer(
            self.trans._render_image, alpha, split_cells=split_cells
        )

    def test_is_on_kitty(self):
        for bg, r in (((0,) * 3, 1), ((100,) * 3, 101), ((255,) * 3, 254), (None, 0)):
            set_fg_bg_colors(bg=bg)
            bg_hex = get_fg_bg_colors(hex=True)[1]
            bg = bg or (0, 0, 0)

            assert all(
                line == SGR_BG_RGB % bg + " " * self.trans.width + SGR_NORMAL
                for line in self.render_image("#").splitlines()
            )
            assert all(
                line == SGR_BG_RGB % bg + " " * self.trans.width + SGR_NORMAL
                for line in self.render_image(bg_hex).splitlines()
            )

            toggle_is_on_kitty()
            bg = (r, *bg[1:])

            assert all(
                line == SGR_BG_RGB % bg + " " * self.trans.width + SGR_NORMAL
                for line in self.render_image("#").splitlines()
            )
            assert all(
                line == SGR_BG_RGB % bg + " " * self.trans.width + SGR_NORMAL
                for line in self.render_image(bg_hex).splitlines()
            )

            toggle_is_on_kitty()
        set_fg_bg_colors((0, 0, 0), (0, 0, 0))

    def test_split_cells(self):
        render_no_split = self.render_image(_ALPHA_THRESHOLD, split_cells=False)
        render_split = self.render_image(_ALPHA_THRESHOLD, split_cells=True)

        assert render_no_split == str(self.trans)
        assert render_split != render_no_split
        assert render_split.replace("\0", "") == render_no_split
        assert render_split.count("\n") == render_no_split.count("\n")

        for split_line, no_split_line in zip(
            render_split.splitlines(), render_no_split.splitlines()
        ):
            cells = split_line.strip(SGR_NORMAL).split("\0")

            assert "".join(split_line.split("\0")) == no_split_line
            assert len(cells) == _size
            for cell in cells:
                assert cell == " "

        python_image = ImageClass(python_img, width=_size)
        render = python_image._renderer(
            python_image._render_image, _ALPHA_THRESHOLD, split_cells=True
        )

        for line, str_line in zip(render.splitlines(), str(python_image).splitlines()):
            cells = line.split("\0")
            assert len(cells) == _size
            assert "".join(cells) == str_line


def test_render_clean_up_All():
    # PIL_IMAGE
    img = Image.open("tests/images/python.png")
    img_copy = img.copy()
    image = ImageClass(img)
    # Source
    image._render_image(img, None)
    img.load()
    # Frame
    image._render_image(img_copy, None, frame=True)
    img_copy.load()
    # Not source and not frame
    image._render_image(img_copy, None)
    with pytest.raises(ValueError, match="closed"):
        img_copy.load()

    # FILE, also applies for URL
    image = ImageClass.from_file("tests/images/python.png")
    img = image._get_image()
    # Frame
    image._render_image(img, None, frame=True)
    img.load()
    # Not source and not frame
    image._render_image(img, None)
    with pytest.raises(ValueError, match="closed"):
        img.load()


def test_style_args_All():
    image = ImageClass(python_img)
    with pytest.raises(StyleError):
        image.draw(_=None)


def test_style_format_spec_All():
    image = ImageClass(python_img)
    for spec in ("+\t", "20+\r", ".^+\a", "#+\0"):
        with pytest.raises(StyleError):
            format(image, spec)


__all__ = [name for name in globals() if name.startswith(("test_", "Test"))]



# Configuration file for the Sphinx documentation builder.
#
# For the full list of configuration options, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

from term_image import __version__, utils
from term_image.image.common import ImageMeta
from term_image.image.iterm2 import ITerm2ImageMeta

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
# sys.path.insert(0, os.path.abspath("../../src"))

# -- Project information -----------------------------------------------------
project = "Term-Image"
copyright = "2022, Toluwaleke Ogundipe"
author = "Toluwaleke Ogundipe"
release = __version__

# -- General configuration ---------------------------------------------------
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx_toolbox.github",
    "sphinx_toolbox.sidebar_links",
    "sphinx_toolbox.more_autosummary",
]

# -- Options for HTML output -------------------------------------------------
html_theme = "furo"
html_logo = "resources/logo.png"
html_favicon = "resources/logo.ico"

# -- Options for extensions ----------------------------------------------

# # -- sphinx-autodoc -----------------------------------------------------
autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
    "member-order": "bysource",
    "autosummary": True,
    "autosummary-nosignatures": True,
}
autodoc_typehints = "description"
autodoc_typehints_format = "fully-qualified"
autodoc_typehints_description_target = "documented"
autodoc_member_order = "bysource"
autodoc_inherit_docstrings = False

# # -- sphinx-intersphinx ----------------------------------------------
intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "pillow": ("https://pillow.readthedocs.io/en/stable/", None),
    "requests": ("https://requests.readthedocs.io/en/stable/", None),
    "urwid": ("https://urwid.org", None),
}

# # -- sphinx_toolbox-github ----------------------------------------------
github_username = "AnonymouX47"
github_repository = "term-image"

# # -- sphinx_toolbox-more_autosummary ----------------------------------------------
autodocsumm_member_order = "bysource"

# -- Event handlers -------------------------------------------------------------


def autodocssumm_grouper(app, what, name, obj, section, parent):
    from enum import EnumMeta
    from types import FunctionType

    if isinstance(obj, EnumMeta):
        return "Enumerations"
    elif isinstance(obj, type) and issubclass(obj, Warning):
        return "Warnings"
    elif isinstance(parent, type):
        obj = vars(parent)[name.rpartition(".")[2]]
        if isinstance(obj, utils.ClassProperty):
            return "Class Properties"
        elif isinstance(obj, utils.ClassInstanceProperty):
            return "Class/Instance Properties"
        elif isinstance(obj, property):
            return "Instance Properties"
        elif isinstance(obj, FunctionType):
            return "Instance Methods"
        elif isinstance(obj, utils.ClassInstanceMethod):
            return "Class/Instance Methods"
        elif isinstance(obj, classmethod):
            return "Class Methods"
        elif isinstance(obj, staticmethod):
            return "Static Methods"


def setup(app):
    app.connect("autodocsumm-grouper", autodocssumm_grouper)


# -- Extras -----------------------------------------------------------

# The properties defined by the metaclass' would be invoked instead of returning the
# property defined by the class
for meta in (ImageMeta, ITerm2ImageMeta):
    for attr, value in tuple(vars(meta).items()):
        if isinstance(value, utils.ClassPropertyBase):
            delattr(meta, attr)



"""Render-style-independent tests"""

import atexit
import io
import os
import sys
from operator import floordiv, mul
from pathlib import Path

import pytest
from PIL import Image, UnidentifiedImageError

from term_image import set_cell_ratio
from term_image.ctlseqs import ESC
from term_image.exceptions import InvalidSizeError, TermImageError
from term_image.image import BaseImage, BlockImage, ImageIterator, ImageSource, Size
from term_image.image.common import _ALPHA_THRESHOLD

from .. import reset_cell_size_ratio
from .common import _size, columns, lines, python_img, setup_common

python_image = "tests/images/python.png"
python_sym = "tests/images/python_sym.png"  # Symlink to "python.png"
anim_img = Image.open("tests/images/lion.gif")
stdout = io.StringIO()

setup_common(BlockImage)
from .common import _height, _width  # noqa:E402


@atexit.register
def close_imgs():
    anim_img.close()


def clear_stdout():
    stdout.seek(0)
    stdout.truncate()


class BytesPath:
    def __init__(self, path: str) -> None:
        self.path = path

    def __fspath__(self) -> bytes:
        return self.path.encode()


class TestConstructor:
    def test_args(self):
        with pytest.raises(TypeError, match=r"'image'"):
            BlockImage(python_image)

        for size in ((0, 1), (1, 0), (0, 0)):
            with pytest.raises(ValueError, match=r"'image'"):
                BlockImage(Image.new("RGB", size))

        # Ensure size arguments get through to `set_size()`
        with pytest.raises(TypeError, match="'width' and 'height'"):
            BlockImage(python_img, width=1, height=Size.FIT)

    def test_init(self):
        image = BlockImage(python_img)
        assert image._size is Size.FIT
        assert image._source is python_img
        assert image._source_type is ImageSource.PIL_IMAGE
        assert isinstance(image._original_size, tuple)
        assert image._original_size == python_img.size

        image = BlockImage(python_img, width=None)
        assert image._size is Size.FIT
        image = BlockImage(python_img, height=None)
        assert image._size is Size.FIT
        image = BlockImage(python_img, width=None, height=None)
        assert image._size is Size.FIT

        image = BlockImage(python_img, width=_size)
        assert isinstance(image._size, tuple)
        image = BlockImage(python_img, height=_size)
        assert isinstance(image._size, tuple)

        for value in Size:
            image = BlockImage(python_img, width=value)
            assert isinstance(image._size, tuple)
            image = BlockImage(python_img, height=value)
            assert isinstance(image._size, tuple)

        image = BlockImage(python_img)
        assert image._is_animated is False

    def test_init_animated(self):
        image = BlockImage(anim_img)
        assert image._is_animated is True
        assert image._frame_duration == (anim_img.info.get("duration") or 100) / 1000
        assert image._seek_position == 0
        assert image._n_frames is None

        try:
            anim_img.seek(2)
            assert BlockImage(anim_img)._seek_position == 2
        finally:
            anim_img.seek(0)


class TestFromFile:
    def test_args(self):
        with pytest.raises(TypeError, match=r"'filepath'"):
            BlockImage.from_file(python_img)
        with pytest.raises(FileNotFoundError):
            BlockImage.from_file(python_image + "e")
        for dir_path in ("tests", Path("tests"), BytesPath("tests")):
            with pytest.raises(OSError):
                BlockImage.from_file(dir_path)
        with pytest.raises(UnidentifiedImageError):
            BlockImage.from_file("LICENSE")

        # Ensure size arguments get through
        with pytest.raises(TypeError, match="'width' and 'height'"):
            BlockImage.from_file(python_image, width=1, height=Size.FIT)

    def test_filepath(self):
        for path in (python_image, Path(python_image), BytesPath(python_image)):
            image = BlockImage.from_file(path)
            assert isinstance(image, BlockImage)
            assert image.source == os.path.abspath(python_image)
            assert image.source_type is ImageSource.FILE_PATH

    @pytest.mark.skipif(
        not os.path.islink(python_sym),
        reason="The symlink is on or has passed through a platform or filesystem that "
        "doesn't support symlinks",
    )
    def test_symlink(self):
        image = BlockImage.from_file(python_sym)
        assert isinstance(image, BlockImage)
        assert image._source == os.path.abspath(python_sym)
        assert image._source_type is ImageSource.FILE_PATH


class TestProperties:
    def test_closed(self):
        image = BlockImage(python_img)
        assert not image.closed

        with pytest.raises(AttributeError):
            image.closed = True

        image.close()
        assert image.closed

    def test_forced_support(self):
        class A(BaseImage):
            pass

        class B(A):
            pass

        class C(B):
            pass

        instance = BlockImage(python_img)

        for value in (1, 1.0, "1", ()):
            with pytest.raises(TypeError):
                A.forced_support = value

        with pytest.raises(AttributeError):
            del BaseImage.forced_support

        with pytest.raises(AttributeError):
            instance.forced_support = True

        with pytest.raises(AttributeError):
            del instance.forced_support

        assert not A.forced_support
        assert not B.forced_support
        assert not C.forced_support

        A.forced_support = True
        assert A.forced_support
        assert B.forced_support
        assert C.forced_support

        B.forced_support = False
        assert A.forced_support
        assert not B.forced_support
        assert not C.forced_support

        C.forced_support = False
        B.forced_support = True
        assert A.forced_support
        assert B.forced_support
        assert not C.forced_support

    def test_frame_duration(self):
        image = BlockImage(python_img)
        assert image.frame_duration is None
        image.frame_duration = 0.5
        assert image.frame_duration is None

        image = BlockImage(anim_img)
        assert image._frame_duration == (anim_img.info.get("duration") or 100) / 1000

        for duration in (0, 1, "0.1", "1", 0.3j):
            with pytest.raises(TypeError):
                image.frame_duration = duration
        for duration in (0.0, -0.1):
            with pytest.raises(ValueError):
                image.frame_duration = duration

        image.frame_duration = 0.5
        assert image.frame_duration == 0.5

    def test_height(self):
        image = BlockImage(python_img)
        assert image.height is Size.FIT

        image.height = _size
        assert isinstance(image.size, tuple)
        assert isinstance(image.width, int)
        assert isinstance(image.height, int)
        assert image.size[1] == _size == image.height

        for value in Size:
            image.height = value
            assert isinstance(image.size, tuple)
            assert isinstance(image.width, int)
            assert isinstance(image.height, int)

        image.height = None
        assert isinstance(image.size, tuple)
        assert isinstance(image.width, int)
        assert isinstance(image.height, int)
        size = image.size
        image.height = Size.FIT
        assert image.size == size

        for value in (0.1, "1", (1, 1), [1, 1]):
            with pytest.raises(TypeError):
                image.height = value
        for value in (-1, 0):
            with pytest.raises(ValueError):
                image.height = value

    def test_is_animated(self):
        image = BlockImage(python_img)
        assert not image.is_animated

        image = BlockImage(anim_img)
        assert image.is_animated

        with pytest.raises(AttributeError):
            image.is_animated = False

    def test_n_frames(self):
        image = BlockImage(python_img)
        assert image.n_frames == 1

        image = BlockImage(anim_img)
        assert 1 < image.n_frames == anim_img.n_frames
        assert image.n_frames == anim_img.n_frames  # Ensure consistency

        with pytest.raises(AttributeError):
            image.n_frames = 2

    def test_original_size(self):
        image = BlockImage(python_img)

        for value in (None, 0, 1, 0.1, "1", (1, 1), [1, 1]):
            with pytest.raises(AttributeError):
                image.original_size = value

        assert image.original_size == python_img.size
        assert isinstance(image.original_size, tuple)

    @reset_cell_size_ratio()
    def test_rendered_size_height_width(self):
        image = BlockImage(python_img)  # Square

        with pytest.raises(AttributeError):
            image.rendered_size = (_size,) * 2
        with pytest.raises(AttributeError):
            image.rendered_height = _size
        with pytest.raises(AttributeError):
            image.rendered_width = _size

        assert isinstance(image.rendered_size, tuple)
        assert isinstance(image.rendered_width, int)
        assert isinstance(image.rendered_height, int)

        image.width = _width

        assert isinstance(image.rendered_size, tuple)
        assert isinstance(image.rendered_width, int)
        assert isinstance(image.rendered_height, int)

        # The rendered size is independent of the cell ratio
        # Change in cell-ratio must not affect the image's rendered size
        set_cell_ratio(0.5)
        image.width = _width
        assert image.rendered_size == (_width, _height)
        set_cell_ratio(0.1)
        assert image.rendered_size == (_width, _height)

    def test_size(self):
        image = BlockImage(python_img)
        assert image.size is Size.FIT

        for value in (None, 0, 1, 0.1, "1", [1, 1]):
            with pytest.raises(TypeError):
                image.size = value

        for value in Size:
            image.size = value
            assert image.size is image.height is image.width is value

        for size in ((1, 1), (100, 50), (50, 100)):
            image.size = size
            assert image.size == size

    def test_source(self):
        image = BlockImage(python_img)
        assert image.source is python_img
        assert image.source_type is ImageSource.PIL_IMAGE

        for path in (python_image, Path(python_image), BytesPath(python_image)):
            image = BlockImage.from_file(path)
            assert image.source == os.path.abspath(python_image)
            assert image.source_type is ImageSource.FILE_PATH

        with pytest.raises(AttributeError):
            image.source = None

    @pytest.mark.skipif(
        not os.path.islink(python_sym),
        reason="The symlink is on or has passed through a platform or filesystem that "
        "doesn't support symlinks",
    )
    def test_source_symlink(self):
        image = BlockImage.from_file(python_sym)
        assert os.path.basename(image.source) == "python_sym.png"
        assert (
            image.source == os.path.abspath(python_sym) != os.path.realpath(python_sym)
        )
        assert image.source_type is ImageSource.FILE_PATH

    def test_width(self):
        image = BlockImage(python_img)
        assert image.width is Size.FIT

        image.width = _size
        assert isinstance(image.size, tuple)
        assert isinstance(image.width, int)
        assert isinstance(image.height, int)
        assert image.size[0] == _size == image.width

        for value in Size:
            image.width = value
            assert isinstance(image.size, tuple)
            assert isinstance(image.width, int)
            assert isinstance(image.height, int)

        image.width = None
        assert isinstance(image.size, tuple)
        assert isinstance(image.width, int)
        assert isinstance(image.height, int)
        size = image.size
        image.width = Size.FIT
        assert image.size == size

        for value in (0.1, "1", (1, 1), [1, 1]):
            with pytest.raises(TypeError):
                image.width = value
        for value in (-1, 0):
            with pytest.raises(ValueError):
                image.width = value


def test_close():
    image = BlockImage(python_img)
    image.close()
    assert image.closed
    with pytest.raises(AttributeError):
        image._source
    with pytest.raises(TermImageError):
        image.source
    with pytest.raises(TermImageError):
        str(image)
    with pytest.raises(TermImageError):
        format(image)
    with pytest.raises(TermImageError):
        image.draw()


def test_context_management():
    image = BlockImage(python_img)
    with image as image2:
        assert image2 is image
    assert image.closed


def test_iter():
    image = BlockImage(python_img)
    with pytest.raises(ValueError, match="not animated"):
        iter(image)

    anim_image = BlockImage(anim_img)
    image_it = iter(anim_image)
    assert isinstance(image_it, ImageIterator)
    assert image_it._image is anim_image


def test_seek_tell():
    # Non-animated
    image = BlockImage(python_img)
    assert image.tell() == 0
    image.seek(0)
    assert image.tell() == 0
    with pytest.raises(ValueError, match="out of range"):
        image.seek(1)
    assert image.tell() == 0

    # Animated
    image = BlockImage(anim_img)
    assert image.tell() == 0
    n_frames = anim_img.n_frames

    image.seek(2)
    assert image.tell() == 2

    with pytest.raises(ValueError, match="out of range"):
        image.seek(n_frames)
    assert image.tell() == 2

    image.seek(n_frames - 1)
    assert image.tell() == n_frames - 1

    with pytest.raises(ValueError, match="out of range"):
        image.seek(n_frames + 1)
    assert image.tell() == n_frames - 1

    image.seek(0)
    assert image.tell() == 0


class TestSetSize:
    image = BlockImage(python_img)  # Square
    h_image = BlockImage.from_file("tests/images/hori.jpg")  # Horizontally-oriented
    v_image = BlockImage.from_file("tests/images/vert.jpg")  # Vertically-oriented

    def test_args_width_height(self):
        for value in (1.0, "1", (), []):
            with pytest.raises(TypeError, match="'width'"):
                self.image.set_size(value)
            with pytest.raises(TypeError, match="'height'"):
                self.image.set_size(height=value)
        for value in (0, -1, -100):
            with pytest.raises(ValueError, match="'width'"):
                self.image.set_size(value)
            with pytest.raises(ValueError, match="'height'"):
                self.image.set_size(height=value)
        for width, height in ((1, Size.FIT), (Size.FIT, 1), (Size.FIT, Size.FIT)):
            with pytest.raises(TypeError, match="'width' and 'height'"):
                self.image.set_size(width, height)

    def test_args_frame_size(self):
        for value in (None, 1, 1.0, "1", (1.0, 1), (1, 1.0), ("1.0",), [1, 1]):
            with pytest.raises(TypeError, match="'frame_size'"):
                self.image.set_size(frame_size=value)
        for value in ((), (0,), (1,), (1, 1, 1)):
            with pytest.raises(ValueError, match="'frame_size'"):
                self.image.set_size(frame_size=value)

    def test_both_width_height(self):
        for size in ((1, 1), (100, 50), (50, 100)):
            self.image.set_size(*size)
            assert self.image.size == size

    @reset_cell_size_ratio()
    def test_frame_size_absolute(self):
        # Some of these assertions pass only when the cell ratio is about 0.5
        set_cell_ratio(0.5)

        self.image.set_size(frame_size=(100, 50))
        assert self.image.size == (100, 50)

        self.image.set_size(frame_size=(100, 55))
        assert self.image.size == (100, 50)

        self.image.set_size(frame_size=(110, 50))
        assert self.image.size == (100, 50)

        self.h_image.set_size(frame_size=(100, 50))
        assert self.h_image.width == 100

        self.v_image.set_size(frame_size=(100, 50))
        assert self.v_image.height == 50

    def test_frame_size_relative(self):
        self.h_image.set_size()
        assert self.h_image.width == columns

        self.v_image.set_size()
        assert self.v_image.height == lines - 2

        self.h_image.set_size(frame_size=(0, 0))
        assert self.h_image.width == columns

        self.v_image.set_size(frame_size=(0, 0))
        assert self.v_image.height == lines

        self.h_image.set_size(frame_size=(-3, 0))
        assert self.h_image.width == columns - 3

        self.v_image.set_size(frame_size=(0, -5))
        assert self.v_image.height == lines - 5

        self.image.set_size(frame_size=(-columns, -lines))
        assert self.image.size == (1, 1)

        self.image.set_size(frame_size=(-columns * 2, -lines * 2))
        assert self.image.size == (1, 1)

    @reset_cell_size_ratio()
    def test_can_exceed_frame_size(self):
        # These assertions pass only when the cell ratio is about 0.5
        set_cell_ratio(0.5)

        self.image.set_size(width=300, frame_size=(200, 100))
        assert self.image.size == (300, 150)

        self.image.set_size(height=150, frame_size=(200, 100))
        assert self.image.size == (300, 150)


def test_renderer():
    def test(img, *args, **kwargs):
        return img, args, kwargs

    image = BlockImage(python_img)

    for positionals, keywords in (
        ((), {}),
        ((2,), {"d": "dude"}),
        ((1, 2, 3), {"a": 1, "b": 2, "c": 3}),
    ):
        img, args, kwargs = image._renderer(test, *positionals, **keywords)
        assert isinstance(img, Image.Image)
        assert args == positionals
        assert kwargs == keywords


class TestRenderData:
    # Fully transparent image
    # It's easy to predict it's pixel values
    trans = BlockImage.from_file("tests/images/trans.png")
    trans.height = _size

    def get_render_data(self, img=None, alpha=None, **kwargs):
        return self.trans._get_render_data(
            img or self.trans._get_image(), alpha, **kwargs
        )

    def test_alpha(self):

        # float
        for alpha in (0.0, _ALPHA_THRESHOLD, 0.999):
            img, _, a = self.get_render_data(alpha=alpha)
            assert isinstance(img, Image.Image)
            assert img.mode == "RGBA"
            assert all(px == 0 for px in a)

            rgb_img = self.trans._get_image().convert("RGB")
            img, _, a = self.get_render_data(rgb_img, alpha)
            assert isinstance(img, Image.Image)
            assert img.mode == "RGB"
            assert all(px == 255 for px in a)

        # str
        for alpha in ("#ffffff", "#"):
            img, _, a = self.get_render_data(alpha=alpha)
            assert isinstance(img, Image.Image)
            assert img.mode == "RGB"
            assert all(px == 255 for px in a)

            rgb_img = self.trans._get_image().convert("RGB")
            img, _, a = self.get_render_data(rgb_img, alpha)
            assert isinstance(img, Image.Image)
            assert img.mode == "RGB"
            assert all(px == 255 for px in a)

        # None
        img, _, a = self.get_render_data(alpha=None)
        assert isinstance(img, Image.Image)
        assert img.mode == "RGB"
        assert all(px == 255 for px in a)

        rgb_img = self.trans._get_image().convert("RGB")
        img, _, a = self.get_render_data(rgb_img, None)
        assert isinstance(img, Image.Image)
        assert img.mode == "RGB"
        assert all(px == 255 for px in a)

    def test_size(self):
        for alpha in (_ALPHA_THRESHOLD, "#", None):
            render_size = self.trans._get_render_size()
            img, rgb, a = self.get_render_data(alpha=alpha)
            assert isinstance(img, Image.Image)
            assert img.size == render_size
            assert len(rgb) == len(a) == mul(*render_size)

            size = (47, 31)
            img, rgb, a = self.get_render_data(alpha=alpha, size=size)
            assert isinstance(img, Image.Image)
            assert img.size == size
            assert len(rgb) == len(a) == mul(*size)

    def test_pixel_data(self):
        for alpha in (_ALPHA_THRESHOLD, "#", None):
            img, rgb, a = self.get_render_data(alpha=alpha)
            assert isinstance(img, Image.Image)
            assert isinstance(rgb, list)
            assert isinstance(a, list)

            img, rgb, a = self.get_render_data(alpha=alpha, pixel_data=False)
            assert isinstance(img, Image.Image)
            assert rgb is None
            assert rgb is None

    def test_round_alpha(self):
        image = BlockImage(python_img, height=_size)
        img, rgb, a = image._get_render_data(python_img, alpha=_ALPHA_THRESHOLD)
        assert isinstance(img, Image.Image)
        assert img.mode == "RGBA"
        assert any(px not in {0, 255} for px in a)

        img, rounded_rgb, a = image._get_render_data(
            python_img, alpha=_ALPHA_THRESHOLD, round_alpha=True
        )
        assert isinstance(img, Image.Image)
        assert img.mode == "RGBA"
        assert all(px in {0, 255} for px in a)

        assert rgb != rounded_rgb  # Blends rounded rgb with terminal BG

    def test_cleanup(self):
        def test(img, *, frame, fail=False):
            ori_size = image._original_size
            if fail:
                ori_img = img
                img = ori_img.copy()

            # resize (no convert)
            image._get_render_data(
                img, 0.5, size=tuple(map(floordiv, ori_size, (2, 2))), frame=frame
            )
            if fail:
                with pytest.raises(ValueError, match="closed"):
                    img.load()
                img = ori_img.copy()
            else:
                img.load()

            # convert (no resize)
            image._get_render_data(img, None, size=ori_size, frame=frame)
            if fail:
                with pytest.raises(ValueError, match="closed"):
                    img.load()
                img = ori_img.copy()
            else:
                img.load()

            # composite (no resize, no convert)
            for alpha in ("#ffffff", "#"):
                image._get_render_data(img, alpha, size=ori_size, frame=frame)
                if fail:
                    with pytest.raises(ValueError, match="closed"):
                        img.load()
                    img = ori_img.copy()
                else:
                    img.load()

            # alpha blend (no resize, no convert)
            image._get_render_data(
                img, 0.5, size=ori_size, round_alpha=True, frame=frame
            )
            if fail:
                with pytest.raises(ValueError, match="closed"):
                    img.load()
                img = ori_img.copy()
            else:
                img.load()

            # no manipulation (no closes)
            source = img if img is image._source else None
            for img in (img, img.convert("RGB")):
                try:
                    if source:
                        image._source = img
                    image._get_render_data(img, 0.5, size=ori_size, frame=frame)
                    img.load()
                finally:
                    if source:
                        image._source = source

        img = Image.open("tests/images/python.png")
        image = BlockImage(img)
        # Source
        test(img, frame=False)
        # Frame
        test(img.copy(), frame=True)
        # Source & Frame
        test(img, frame=True)
        # Not source & not frame
        test(img.copy(), frame=False, fail=True)


# As long as each subclass passes it's render tests (particularly those related to the
# size of the render results), then testing formatting with a single style should
# suffice.
class TestFormatting:
    image = BlockImage(python_img)
    image._size = (15, 15)
    render = str(image)
    check_formatting = staticmethod(image._check_formatting)

    def check_padding(self, result, *fmt):
        render = self.image._format_render(self.render, *self.check_formatting(*fmt))
        left, right = [], []

        chunk, _, render = render.partition(ESC)
        top, _, first_left = chunk.rpartition("\n")
        left.append(first_left)

        chunk, _, render = render.partition("\n")
        right.append(chunk.rpartition("m")[2])

        render, _, chunk = render.rpartition("m")
        last_right, _, bottom = chunk.partition("\n")

        render, _, chunk = render.rpartition("\n")
        last_left = chunk.partition(ESC)[0]

        for chunk in render.splitlines():
            next_left, _, chunk = chunk.partition(ESC)
            left.append(next_left)
            right.append(chunk.rpartition("m")[2])

        left.append(last_left)
        right.append(last_right)
        top, bottom = top.splitlines(), bottom.splitlines()

        # unquote to debug padding
        """
        print(self.image.size)
        for name in ("left", "right", "top", "bottom"):
            side = vars()[name]
            print(f"------------ {name} - {len(side)} --------------")
            for line in side:
                print(f"{len(line)} {line!r}")
        raise ValueError
        """

        width = fmt[1]
        width = width if width > 0 else max(columns + width, 1)
        n_top, n_left, n_bottom, n_right = result

        assert len(left) == self.image.rendered_height
        assert all(line == " " * n_left for line in left)
        assert len(right) == self.image.rendered_height
        assert all(line == " " * n_right for line in right)
        assert len(top) == n_top
        assert all(line == " " * width for line in top)
        assert len(bottom) == n_bottom
        assert all(line == " " * width for line in bottom)

    def test_args(self):
        for value in (1, 1.0, (), []):
            with pytest.raises(TypeError, match="'h_align'"):
                self.check_formatting(h_align=value)
            with pytest.raises(TypeError, match="'v_align'"):
                self.check_formatting(v_align=value)

        for value in ("", "cool", ".", " ", "\n"):
            with pytest.raises(ValueError, match="'h_align'"):
                self.check_formatting(h_align=value)
            with pytest.raises(ValueError, match="'v_align'"):
                self.check_formatting(v_align=value)

        for value in (None, "1", 1.0, (), []):
            with pytest.raises(TypeError, match="'pad_width'"):
                self.check_formatting(width=value)
            with pytest.raises(TypeError, match="'pad_height'"):
                self.check_formatting(height=value)

    def test_arg_align_conversion(self):
        assert self.check_formatting() == (None, columns, None, lines - 2)

        for value in "<|>":
            assert self.check_formatting(h_align=value)[0] == value
        for val1, val2 in zip(("left", "center", "right"), "<|>"):
            assert self.check_formatting(h_align=val1)[0] == val2

        for value in "^-_":
            assert self.check_formatting(v_align=value)[2] == value
        for val1, val2 in zip(("top", "middle", "bottom"), "^-_"):
            assert self.check_formatting(v_align=val1)[2] == val2

    def test_arg_padding_width_absolute(self):
        for value in (1, _width, columns):
            assert self.check_formatting(width=value)[1] == value

        # Can exceed terminal width
        assert self.check_formatting(width=columns + 1)[1] == columns + 1

    def test_arg_padding_width_relative(self):
        for value in (0, -1, -2, -(columns - 1)):
            assert self.check_formatting(width=value)[1] == columns + value

        # Out of range
        for value in (-columns, -(columns + 1), -(columns * 2)):
            assert self.check_formatting(width=value)[1] == 1

    def test_arg_padding_height_absolute(self):
        for value in (1, _size, lines):
            assert self.check_formatting(height=value)[3] == value

        # Can exceed terminal height
        assert self.check_formatting(height=lines + 1)[3] == lines + 1

    def test_arg_padding_height_relative(self):
        for value in (0, -1, -2, -(lines - 1)):
            assert self.check_formatting(height=value)[3] == lines + value

        # Out of range
        for value in (-lines, -(lines + 1), -(lines * 2)):
            assert self.check_formatting(height=value)[3] == 1

    def test_padding_width_left(self):
        for width, result in (
            (15, (0, 0, 0, 0)),
            (16, (0, 0, 0, 1)),
            (17, (0, 0, 0, 2)),
            (29, (0, 0, 0, 14)),
            (30, (0, 0, 0, 15)),
        ):
            self.check_padding(result, "<", width, None, 15)

    def test_padding_width_center(self):
        for width, result in (
            (15, (0, 0, 0, 0)),
            (16, (0, 0, 0, 1)),
            (17, (0, 1, 0, 1)),
            (29, (0, 7, 0, 7)),
            (30, (0, 7, 0, 8)),
        ):
            self.check_padding(result, "|", width, None, 15)
            self.check_padding(result, None, width, None, 15)

    def test_padding_width_right(self):
        for width, result in (
            (15, (0, 0, 0, 0)),
            (16, (0, 1, 0, 0)),
            (17, (0, 2, 0, 0)),
            (29, (0, 14, 0, 0)),
            (30, (0, 15, 0, 0)),
        ):
            self.check_padding(result, ">", width, None, 15)

    def test_padding_width_relative(self):
        # terminal width = 80
        for relative, width, result in (
            (0, 80, (0, 65, 0, 0)),
            (-1, 79, (0, 64, 0, 0)),
            (-2, 78, (0, 63, 0, 0)),
            (-64, 16, (0, 1, 0, 0)),
            (-65, 15, (0, 0, 0, 0)),
            (-80, 1, (0, 0, 0, 0)),
            (-100, 1, (0, 0, 0, 0)),
        ):
            self.check_padding(result, ">", width, None, 15)
            self.check_padding(result, ">", relative, None, 15)

    def test_padding_height_top(self):
        for height, result in (
            (15, (0, 0, 0, 0)),
            (16, (0, 0, 1, 0)),
            (17, (0, 0, 2, 0)),
            (29, (0, 0, 14, 0)),
            (30, (0, 0, 15, 0)),
        ):
            self.check_padding(result, None, 15, "^", height)

    def test_padding_height_middle(self):
        for height, result in (
            (15, (0, 0, 0, 0)),
            (16, (0, 0, 1, 0)),
            (17, (1, 0, 1, 0)),
            (29, (7, 0, 7, 0)),
            (30, (7, 0, 8, 0)),
        ):
            self.check_padding(result, None, 15, "-", height)
            self.check_padding(result, None, 15, None, height)

    def test_padding_height_bottom(self):
        for height, result in (
            (15, (0, 0, 0, 0)),
            (16, (1, 0, 0, 0)),
            (17, (2, 0, 0, 0)),
            (29, (14, 0, 0, 0)),
            (30, (15, 0, 0, 0)),
        ):
            self.check_padding(result, None, 15, "_", height)

    def test_padding_height_relative(self):
        # terminal height = 30
        for relative, height, result in (
            (0, 30, (15, 0, 0, 0)),
            (-1, 29, (14, 0, 0, 0)),
            (-2, 28, (13, 0, 0, 0)),
            (-14, 16, (1, 0, 0, 0)),
            (-15, 15, (0, 0, 0, 0)),
            (-30, 1, (0, 0, 0, 0)),
            (-50, 1, (0, 0, 0, 0)),
        ):
            self.check_padding(result, None, 15, "_", height)
            self.check_padding(result, None, 15, "_", relative)

    def test_mixed_align(self):
        self.check_padding((0, 0, 15, 15), "<", 30, "^", 30)
        self.check_padding((7, 7, 8, 8), "|", 30, "-", 30)
        self.check_padding((15, 15, 0, 0), ">", 30, "_", 30)

    def test_format_spec(self):
        for spec in (
            "1<",
            "-1.|1",
            "<1.1^",
            ".",
            "1.",
            "<.",
            ">1.",
            "-",
            "<^",
            ".#",
            ">1.#.23",
            "#0",
            "#.",
            "#2445",
            "#.23fa45",
            "#fffffff",
            "#a45gh4",
            "###",
            " ",
            # style-specific section
            "+",
            "20+",
            ".^+",
            "#+",
        ):
            with pytest.raises(ValueError, match=r"Invalid format specifier"):
                self.image._check_format_spec(spec)

        for spec in (
            "<",
            "1",
            ".1",
            "<1",
            ".^1",
            "|1.-1",
            "<1.-",
            ".-",
            "#",
            "##",
            "#123456",
            "#23af5b",
            "#23AF5B",
            "#23Af5B",
            "#abcdef",
            "#ABCDEF",
            "#AbCdEf",
            "#.4",
            "#.343545453453",
            "1.1#",
            "1.1##",
            "<.^#ffffff",
            "<.^#FFFFFF",
            "<.^#fFfFfF",
            "<1.^1#.2",
            f"<{columns}.^{lines}##",
        ):
            fmt = self.image._check_format_spec(spec)
            assert isinstance(fmt, tuple)


# Testing with one style should suffice for all since it's simply testing the method
# and nothing perculiar to the style
class TestDraw:
    image = BlockImage(python_img, width=_size)
    anim_image = BlockImage(anim_img, width=_size)

    def test_args(self):
        sys.stdout = stdout
        for value in (1, (), [], {}, b""):
            with pytest.raises(TypeError, match="'alpha'"):
                self.image.draw(alpha=value)

        for value in (-1.0, -0.1, 1.0, 1.1):
            with pytest.raises(ValueError, match="'alpha'"):
                self.image.draw(alpha=value)

        for value in ("f", "fffff", "fffffff", "12h45g", "-2343"):
            with pytest.raises(ValueError, match="Invalid hex color"):
                self.image.draw(alpha=value)

        with pytest.raises(ValueError, match="'pad_width'"):
            self.image.draw(pad_width=columns + 1)

        with pytest.raises(ValueError, match="'pad_height'"):
            self.anim_image.draw(pad_height=lines + 1)

        for value in (1, 1.0, "1", (), []):
            with pytest.raises(TypeError, match="'animate'"):
                self.anim_image.draw(animate=value)

        for arg in ("scroll", "check_size"):
            for value in (1, 1.0, "1", (), []):
                with pytest.raises(TypeError, match=f"{arg!r}"):
                    self.image.draw(**{arg: value})

    def test_size_validation(self):
        sys.stdout = stdout
        self.image._size = (columns + 1, 1)
        with pytest.raises(InvalidSizeError, match="image cannot .* terminal size"):
            self.image.draw()

        self.image._size = (1, lines + 1)
        with pytest.raises(InvalidSizeError, match="image cannot .* terminal size"):
            self.image.draw()

    class TestNonAnimated:
        image = BlockImage(python_img, width=_size)
        anim_image = BlockImage(anim_img, width=_size)

        def test_fit_to_width_width(self):
            sys.stdout = stdout
            self.image._size = (columns, lines + 1)
            with pytest.raises(InvalidSizeError, match="image cannot .* terminal size"):
                self.image.draw()

        def test_fit_to_width_height(self):
            sys.stdout = stdout
            self.image._size = (columns, lines + 1)
            with pytest.raises(InvalidSizeError, match="image cannot .* terminal size"):
                self.image.draw()

        def test_scroll(self):
            sys.stdout = stdout
            self.image._size = (columns, lines + 1)
            self.image.draw(scroll=True)
            assert stdout.getvalue().count("\n") == lines + 1
            clear_stdout()

        def test_check_size(self):
            sys.stdout = stdout
            self.image._size = (columns + 1, lines)
            self.image.draw(check_size=False)
            assert stdout.getvalue().count("\n") == lines
            clear_stdout()

    class TestAnimatedFalse:
        image = BlockImage(python_img, width=_size)
        anim_image = BlockImage(anim_img, width=_size)

        def test_fit_to_width_width(self):
            sys.stdout = stdout
            self.anim_image._size = (columns, lines + 1)
            with pytest.raises(InvalidSizeError, match="image cannot .* terminal size"):
                self.anim_image.draw(animate=False)

        def test_fit_to_width_height(self):
            sys.stdout = stdout
            self.anim_image._size = (columns, lines + 1)
            with pytest.raises(InvalidSizeError, match="image cannot .* terminal size"):
                self.anim_image.draw(animate=False)

        def test_scroll(self):
            sys.stdout = stdout
            self.anim_image._size = (columns, lines + 1)
            self.anim_image.draw(scroll=True, animate=False)
            assert stdout.getvalue().count("\n") == lines + 1
            clear_stdout()

        def test_check_size(self):
            sys.stdout = stdout
            self.anim_image._size = (columns + 1, lines)
            self.anim_image.draw(animate=False, check_size=False)
            assert stdout.getvalue().count("\n") == lines
            clear_stdout()

    class TestAnimated:
        image = BlockImage(python_img)
        anim_image = BlockImage(anim_img)

        def test_fit_to_width_width(self):
            sys.stdout = stdout
            self.anim_image._size = (columns, lines + 1)
            with pytest.raises(
                InvalidSizeError, match="animation cannot .* terminal size"
            ):
                self.anim_image.draw()

        def test_fit_to_width_height(self):
            sys.stdout = stdout
            self.anim_image._size = (columns, lines + 1)
            with pytest.raises(
                InvalidSizeError, match="animation cannot .* terminal size"
            ):
                self.anim_image.draw()

        def test_scroll(self):
            sys.stdout = stdout
            self.anim_image._size = (columns, lines + 1)
            with pytest.raises(InvalidSizeError, match="rendered height .* animation"):
                self.anim_image.draw(scroll=True)

        def test_check_size(self):
            sys.stdout = stdout
            self.anim_image._size = (columns + 1, lines)
            with pytest.raises(
                InvalidSizeError, match="animation cannot .* terminal size"
            ):
                self.anim_image.draw(check_size=False)

        def test_fit_scroll_check_size(self):
            sys.stdout = stdout
            self.anim_image._size = (columns + 1, lines + 1)
            with pytest.raises(
                InvalidSizeError, match="animation cannot .* terminal size"
            ):
                self.anim_image.draw(scroll=True, check_size=False)



import atexit
from types import GeneratorType

import pytest
from PIL import Image

from term_image.ctlseqs import SGR_NORMAL
from term_image.exceptions import TermImageError
from term_image.image import BlockImage, ImageIterator, Size

_size = (30, 15)

png_img = Image.open("tests/images/python.png")
png_image = BlockImage(png_img)
gif_img = Image.open("tests/images/lion.gif")
gif_image = BlockImage(gif_img)
webp_img = Image.open("tests/images/anim.webp")
webp_image = BlockImage(webp_img)

gif_image._size = _size
webp_image._size = _size


@atexit.register
def close_imgs():
    for img in (png_img, gif_img, webp_img):
        img.close()


def test_args():
    for value in ("tests/images/anim.webp", gif_img, webp_img):
        with pytest.raises(TypeError, match="'image'"):
            ImageIterator(value)
    with pytest.raises(ValueError, match="not animated"):
        ImageIterator(png_image)

    for value in (None, 2.0, 0.2, "2"):
        with pytest.raises(TypeError, match="'repeat'"):
            ImageIterator(gif_image, value)
    with pytest.raises(ValueError, match="'repeat'"):
        ImageIterator(gif_image, 0)

    for value in (None, 2.0, 2):
        with pytest.raises(TypeError, match="'format_spec'"):
            ImageIterator(gif_image, format_spec=value)
    with pytest.raises(ValueError, match="format specifier"):
        ImageIterator(gif_image, format_spec=".")

    for value in (None, 2.0, "2"):
        with pytest.raises(TypeError, match="'cached'"):
            ImageIterator(gif_image, cached=value)
    for value in (0, -1, -10):
        with pytest.raises(ValueError, match="'cached'"):
            ImageIterator(gif_image, cached=value)


class TestInit:
    def test_defaults(self):
        for image in (gif_image, webp_image):
            image_it = ImageIterator(image)
            assert image_it._image is image
            assert image_it._repeat == -1
            assert image_it._format == ""
            assert image_it._cached is (image.n_frames <= 100)
            assert isinstance(image_it._animator, GeneratorType)

    def test_with_args(self):
        for repeat, format_spec, cached in (
            (-1, "", 100),
            (2, "#", True),
            (10, "1.1", False),
            (100, "#.9", 1),
        ):
            image_it = ImageIterator(gif_image, repeat, format_spec, cached)
            assert image_it._image is gif_image
            assert image_it._repeat == repeat
            assert image_it._format == format_spec
            assert image_it._cached is (
                cached if isinstance(cached, bool) else gif_image.n_frames <= cached
            )
            assert isinstance(image_it._animator, GeneratorType)

    def test_no_caching_if_repeat_equals_1(self):
        for value in (True, 1, 100):
            image_it = ImageIterator(gif_image, 1, cached=value)
            assert image_it._cached is False

    def test_image_seek_position_unchanged(self):
        gif_image.seek(2)
        image_it = ImageIterator(gif_image)
        assert gif_image.tell() == 2

        next(image_it)
        assert gif_image.tell() == 0


def test_next():
    image_it = ImageIterator(gif_image, 1, "1.1")
    assert isinstance(next(image_it), str)

    for _ in range(gif_image.n_frames - 1):
        next(image_it)

    with pytest.raises(StopIteration):
        next(image_it)

    # Frame number is set to zero
    assert gif_image.tell() == 0

    # Iterator is closed
    assert not hasattr(image_it, "_animator")
    assert not hasattr(image_it, "_img")

    # All calls after StopIteration is first raised also raise StopIteration
    for _ in range(2):
        with pytest.raises(StopIteration):
            next(image_it)


def test_image_seek_has_no_effect():
    image_it = ImageIterator(gif_image, 1)
    next(image_it)
    assert gif_image.tell() == 0

    gif_image.seek(4)
    next(image_it)
    assert gif_image.tell() == 1


def test_iter():
    image_it = ImageIterator(gif_image, 1, "1.1")
    assert iter(image_it) is image_it

    # Image seek position is updated
    for n, _ in enumerate(ImageIterator(gif_image, 1, "1.1")):
        assert gif_image.tell() == n

    for image in (gif_image, webp_image):
        frames = tuple(ImageIterator(image, 1, "1.1"))
        assert len(frames) == image.n_frames
        assert all(isinstance(frame, str) for frame in frames)

        # Consecutive frames are different
        prev_frame = None
        for frame in frames:
            assert frame != prev_frame
            prev_frame = frame

    # Frames are the same as for manual iteration
    gif_image2 = BlockImage(gif_img)  # Need to change image size
    gif_image2._size = _size
    for n, frame in enumerate(ImageIterator(gif_image, 1, "1.1")):
        gif_image2.seek(n)
        assert frame == str(gif_image2)


def test_repeat():
    for image in (gif_image, webp_image):
        for value in (False, True):
            frames = tuple(ImageIterator(image, 2, "1.1", cached=value))

            # # Number of frames is multiplied
            assert len(frames) == image.n_frames * 2

            # # Corresponding frames in different repeat loops are the same
            assert frames[: image.n_frames] == frames[image.n_frames :]


def test_caching():
    def render(*args, **kwargs):
        nonlocal n_calls
        n_calls += 1
        if gif_image2.tell() == gif_image2.n_frames:
            raise EOFError
        return ""

    gif_image2 = BlockImage(gif_img)  # Need to change image size
    gif_image2._size = _size
    gif_image2._render_image = render

    n_calls = 0
    [*ImageIterator(gif_image2, 2, "1.1", cached=True)]
    assert n_calls == gif_image2.n_frames + 1  # +1 for EOF call

    n_calls = 0
    [*ImageIterator(gif_image2, 2, "1.1", cached=False)]
    assert n_calls == gif_image2.n_frames * 2 + 2  # +2 for EOF calls

    del gif_image2._render_image
    image_it = ImageIterator(gif_image2, 4, "1.1", cached=True)

    gif_image2._size = (20, 40)
    frame_0_1 = next(image_it)
    assert frame_0_1.count("\n") + 1 == 40
    image_it.seek(gif_image2.n_frames - 1)
    assert next(image_it).count("\n") + 1 == 40

    # Unchanged
    assert next(image_it) is frame_0_1
    image_it.seek(gif_image2.n_frames - 1)
    assert next(image_it).count("\n") + 1 == 40

    # Change in size
    gif_image2._size = (40, 20)
    frame_0_2 = next(image_it)
    assert frame_0_2 is not frame_0_1
    assert frame_0_2.count("\n") + 1 == 20
    image_it.seek(gif_image2.n_frames - 1)
    assert next(image_it).count("\n") + 1 == 20

    image_it.close()


def test_sizing():
    def test(image_it):
        for value in Size:
            gif_image2.size = value
            assert next(image_it).count("\n") + 1 == gif_image2.rendered_height
            assert gif_image2.size is value

        gif_image2._size = (40, 20)
        assert next(image_it).count("\n") + 1 == 20
        assert gif_image2._size == (40, 20)

        gif_image2._size = (20, 10)
        assert next(image_it).count("\n") + 1 == 10
        assert gif_image2._size == (20, 10)

        gif_image2.size = Size.FIT
        next(image_it)
        assert gif_image2.size is Size.FIT

    gif_image2 = BlockImage(gif_img)  # Need to change image size

    # Uncached loop
    image_it = ImageIterator(gif_image2, 1, "1.1")
    test(image_it)

    # Cached loop
    image_it = ImageIterator(gif_image2, 2, "1.1", True)
    for _ in range(gif_image2.n_frames):
        next(image_it)
    test(image_it)


def test_formatting():
    # Transparency enabled, not padded
    image_it = ImageIterator(gif_image, 1, "1.1")
    assert next(image_it).count("\n") + 1 == _size[1]
    # First line without escape codes
    assert next(image_it).partition("\n")[0].strip(SGR_NORMAL) == " " * _size[0]

    # Transparency disabled, not padded
    image_it = ImageIterator(gif_image, 1, "1.1#")
    assert next(image_it).count("\n") + 1 == _size[1]
    # First line without escape codes
    assert next(image_it).partition("\n")[0].strip(SGR_NORMAL) != " " * _size[0]

    # Transparency disabled, padded
    image_it = ImageIterator(gif_image, 1, f"{_size[0] + 2}.{_size[1] + 2}#")
    assert next(image_it).count("\n") + 1 == _size[1] + 2
    # First line should be padding, so no escape codes
    assert next(image_it).partition("\n")[0] == " " * (_size[0] + 2)


def test_loop_no():
    for cached in (False, True):
        image_it = ImageIterator(gif_image, 2, cached=cached)
        assert image_it.loop_no is None

        next(image_it)
        assert image_it.loop_no == 2
        for _ in range(gif_image.n_frames - 1):
            next(image_it)
        assert image_it.loop_no == 2

        next(image_it)
        assert image_it.loop_no == 1
        for _ in range(gif_image.n_frames - 1):
            next(image_it)
        assert image_it.loop_no == 1

        with pytest.raises(StopIteration):
            next(image_it)
        assert image_it.loop_no == 0


def test_close():
    # From PIL image-sourced image
    image_it = ImageIterator(gif_image, 1)
    next(image_it)
    img = image_it._img
    assert img is gif_img

    image_it.close()
    assert gif_image.tell() == 0
    assert not hasattr(image_it, "_animator")
    assert not hasattr(image_it, "_img")
    assert img.load()

    # File-sourced image
    image_it = ImageIterator(BlockImage.from_file(gif_image._source.filename), 1)
    next(image_it)
    img = image_it._img
    image_it.close()
    assert gif_image.tell() == 0
    assert not hasattr(image_it, "_animator")
    assert not hasattr(image_it, "_img")
    with pytest.raises(ValueError, match="Operation on closed image"):
        img.load()


class TestSeek:
    def test_seek_args(self):
        image_it = ImageIterator(gif_image)

        for value in (1.0, "1", [], ()):
            with pytest.raises(TypeError):
                image_it.seek(value)

        for value in (-2, -1, gif_image.n_frames, gif_image.n_frames + 1):
            with pytest.raises(ValueError):
                image_it.seek(value)

        with pytest.raises(TermImageError):
            image_it.seek(1)

    def test_seek(self):
        for cached in (False, True):
            image_it = ImageIterator(gif_image, 2, cached=cached)
            frame_0 = next(image_it)
            for _ in range(5):
                next(image_it)
            frame_6 = next(image_it)

            # Image seek position is always the number of the last yielded frame. Hence:
            # - must remain unchanged after `seek()`.
            # - must equal the seek-ed position after the `next()` immediately following
            # `seek()`.

            # 1st loop

            image_it.seek(0)
            assert image_it._loop_no == 2
            assert gif_image.tell() == 6
            assert frame_0 == next(image_it)
            assert image_it._loop_no == 2
            assert gif_image.tell() == 0

            image_it.seek(6)
            assert image_it._loop_no == 2
            assert gif_image.tell() == 0
            assert frame_6 == next(image_it)
            assert image_it._loop_no == 2
            assert gif_image.tell() == 6

            image_it.seek(gif_image.n_frames - 1)
            assert image_it._loop_no == 2
            assert gif_image.tell() == 6
            next(image_it)
            assert image_it._loop_no == 2
            assert gif_image.tell() == gif_image.n_frames - 1

            # 2nd loop (cached when `cached=True`)

            assert frame_0 == next(image_it)
            assert image_it._loop_no == 1
            assert gif_image.tell() == 0

            image_it.seek(6)
            assert image_it._loop_no == 1
            assert gif_image.tell() == 0
            assert frame_6 == next(image_it)
            assert image_it._loop_no == 1
            assert gif_image.tell() == 6

            image_it.seek(0)
            assert image_it._loop_no == 1
            assert gif_image.tell() == 6
            assert frame_0 == next(image_it)
            assert image_it._loop_no == 1
            assert gif_image.tell() == 0

            image_it.seek(gif_image.n_frames - 1)
            assert image_it._loop_no == 1
            assert gif_image.tell() == 0
            next(image_it)
            assert image_it._loop_no == 1
            assert gif_image.tell() == gif_image.n_frames - 1

            # Iteration ends normally
            with pytest.raises(StopIteration):
                next(image_it)

            # After closing
            with pytest.raises(TermImageError):
                image_it.seek(0)

    def test_seek_skipped_frames_caching(self):
        image_it = ImageIterator(gif_image, 1)
        next(image_it)
        frame_1 = next(image_it)
        for _ in range(4):
            next(image_it)
        frame_6 = next(image_it)

        image_it = ImageIterator(gif_image, 2, cached=True)
        next(image_it)
        image_it.seek(gif_image.n_frames - 1)
        next(image_it) and next(image_it)
        cache = image_it._animator.gi_frame.f_locals["cache"]

        assert frame_1 == next(image_it) is cache[1][0]
        image_it.seek(6)
        assert frame_6 == next(image_it) is cache[6][0]
